# -*- coding: utf-8 -*-
"""
Test the functions related to configuring proposal methods
"""
import numpy as np
import pytest
from unittest.mock import MagicMock

from nessai.samplers.nestedsampler import NestedSampler
from nessai.proposal import (
    AnalyticProposal,
    AugmentedFlowProposal,
    RejectionProposal,
    FlowProposal,
)
from nessai.gw.proposal import (
    AugmentedGWFlowProposal,
    GWFlowProposal,
)


@pytest.fixture
def sampler(sampler, tmpdir):
    sampler.output = tmpdir.mkdir("test")
    sampler.plot = False
    sampler.n_pool = None
    sampler.acceptance_threshold = 0.01
    return sampler


@pytest.mark.parametrize(
    "maximum, result", [[None, 200], [False, 0], [100, 100], ["inf", np.inf]]
)
def test_uninformed_maximum(sampler, maximum, result):
    """
    Test to check that the proposal is correctly configured depending on
    the maximum number of uninformed iterations.
    """
    NestedSampler.configure_uninformed_proposal(
        sampler, None, False, maximum, None
    )
    assert sampler.maximum_uninformed == result

    if maximum is False:
        assert sampler.uninformed_sampling is False
    else:
        assert sampler.uninformed_sampling is True


def test_uninformed_threshold(sampler):
    """Test the check uninformed threshold is set correctly"""
    NestedSampler.configure_uninformed_proposal(
        sampler, None, False, None, 0.5
    )
    assert sampler.uninformed_acceptance_threshold == 0.5


def test_uninformed_threshold_default_below(sampler):
    """
    Test to check that the threshold is set to 10 times the acceptance
    if it is below 0.1.
    """
    sampler.acceptance_threshold = 0.05
    NestedSampler.configure_uninformed_proposal(
        sampler, None, False, None, None
    )
    assert sampler.uninformed_acceptance_threshold == 0.5


@pytest.mark.parametrize("threshold", [0.1, 0.2])
def test_uninformed_threshold_default_(sampler, threshold):
    """
    Test to check that the threshold is set to the same value if it is above
    or equal to 0.1
    """
    sampler.acceptance_threshold = threshold
    NestedSampler.configure_uninformed_proposal(
        sampler, None, False, None, None
    )
    assert sampler.uninformed_acceptance_threshold == threshold


def test_uninformed_no_analytic_priors(sampler):
    """
    Test to check that the correct proposal method is used without analytic
    priors.
    """
    NestedSampler.configure_uninformed_proposal(
        sampler, None, False, None, None
    )
    assert isinstance(sampler._uninformed_proposal, RejectionProposal)


def test_uninformed_analytic_priors(sampler):
    """
    Test to check that the correct proposal method is used with analytic
    priors.
    """
    NestedSampler.configure_uninformed_proposal(
        sampler, None, True, None, None
    )
    assert isinstance(sampler._uninformed_proposal, AnalyticProposal)


def test_uninformed_proposal_class(sampler):
    """Test using a custom proposal class"""
    from nessai.proposal.base import Proposal

    class TestProposal(Proposal):
        def draw(self, point):
            pass

    NestedSampler.configure_uninformed_proposal(
        sampler, TestProposal, False, None, None
    )
    assert isinstance(sampler._uninformed_proposal, TestProposal)


def test_no_flow_proposal_class(sampler):
    """Test the default flow class"""
    NestedSampler.configure_flow_proposal(sampler, None, {}, False)
    assert isinstance(sampler._flow_proposal, FlowProposal)


@pytest.mark.parametrize(
    "flow_class, result_class",
    [
        ["FlowProposal", FlowProposal],
        ["AugmentedFlowProposal", AugmentedFlowProposal],
        ["GWFlowProposal", GWFlowProposal],
        ["AugmentedGWFlowProposal", AugmentedGWFlowProposal],
        ["flowproposal", FlowProposal],
        ["augmentedflowproposal", AugmentedFlowProposal],
        ["gwflowproposal", GWFlowProposal],
        ["augmentedgwflowproposal", AugmentedGWFlowProposal],
    ],
)
def test_flow__class(flow_class, result_class, sampler):
    """Test the correct class is imported and used"""
    NestedSampler.configure_flow_proposal(sampler, flow_class, {}, False)
    assert isinstance(sampler._flow_proposal, result_class)


def test_unknown_flow_class(sampler):
    """Test to check the error raised if an unknown class is used"""
    with pytest.raises(ValueError) as excinfo:
        NestedSampler.configure_flow_proposal(sampler, "GWProposal", {}, False)
    assert "Unknown flow class" in str(excinfo.value)


def test_flow_class_not_subclass(sampler):
    """
    Test to check an error is raised in the class does not inherit from
    FlowProposal
    """

    class FakeProposal:
        pass

    with pytest.raises(RuntimeError) as excinfo:
        NestedSampler.configure_flow_proposal(sampler, FakeProposal, {}, False)
    assert "inherits" in str(excinfo.value)


@pytest.mark.parametrize("val", [(0.1, 10), (0.8, 200)])
def test_proposal_switch(sampler, val):
    """Test the method for switching proposals"""
    sampler.mean_block_acceptance = 0.5
    sampler.mean_acceptance = val[0]
    sampler.uninformed_acceptance_threshold = 0.5
    sampler.iteration = val[1]
    sampler.maximum_uninformed = 100
    sampler._flow_proposal = MagicMock()
    sampler._flow_proposal.n_pool = 2
    sampler._flow_proposal.configure_pool = MagicMock()
    sampler._uninformed_proposal = MagicMock()
    sampler._uninformed_proposal.pool = True
    sampler._uninformed_proposal.close_pool = MagicMock()
    sampler.proposal = sampler._uninformed_proposal

    assert NestedSampler.check_proposal_switch(sampler) is True

    assert sampler.uninformed_sampling is False
    assert sampler.proposal == sampler._flow_proposal


def test_proposal_no_switch(sampler):
    """Ensure proposal is not switched"""
    sampler.mean_acceptance = 0.5
    sampler.uninformed_acceptance_threshold = 0.1
    sampler.iteration = 10
    sampler.maximum_uninformed = 100
    assert NestedSampler.check_proposal_switch(sampler) is False


def test_proposal_already_switched(sampler):
    """Test switching when proposal is already switched"""
    sampler.mean_acceptance = 0.5
    sampler.uninformed_acceptance_threshold = 0.1
    sampler.mean_block_acceptance = 0.5
    sampler.iteration = 10
    sampler.maximum_uninformed = 100
    sampler._flow_proposal = MagicMock()
    sampler._flow_proposal.ns_acceptance = 0.2
    sampler.uninformed_sampling = False
    sampler.proposal = sampler._flow_proposal
    assert NestedSampler.check_proposal_switch(sampler, force=True) is True
    assert sampler.proposal is sampler._flow_proposal
    # If proposal was switched again, acceptance would change
    assert sampler.proposal.ns_acceptance == 0.2



# -*- coding: utf-8 -*-
"""Utilities for proposal classes"""
from copy import deepcopy
from inspect import getmro, signature
import logging


logger = logging.getLogger(__name__)


def check_proposal_kwargs(ProposalClass, kwargs, strict=False):
    """Check the keyword arguments are correct for a proposal class.

    Removes any keyword arguments that correspond to another class. Raises an
    error if the keyword arguments are unknown. For example if the class is
    FlowProposal, it will remove any keyword arguments that correspond to
    AugmentedFlowProposal but not raise an error.

    Parameters
    ----------
    ProposalClass : Type[FlowProposal]
        Class to check the keyword arguments against.
    kwargs : dict
        Dictionary of keyword arguments.
    strict : bool
        Raise an error if any extra keyword arguments are provided even if they
        correspond to other proposals.

    Returns
    -------
    dict
        Dictionary of updated kwargs.
    """
    from ..proposal import AugmentedFlowProposal, FlowProposal
    from ..gw.proposal import AugmentedGWFlowProposal, GWFlowProposal

    proposals = {
        AugmentedFlowProposal,
        AugmentedGWFlowProposal,
        FlowProposal,
        GWFlowProposal,
    }

    class_keys = set()
    for cls in getmro(ProposalClass):
        class_keys.update(signature(cls).parameters.keys())
        if cls in proposals:
            proposals.remove(cls)

    keys = set(kwargs.keys())
    kwargs_out = deepcopy(kwargs)

    extra_keys = keys - class_keys

    if not extra_keys:
        logger.debug("All keyword arguments match the proposal class")
        return kwargs_out
    elif strict:
        raise RuntimeError(
            f"Keyword arguments contain unknown keys: {extra_keys}"
        )

    allowed_extra_keys = set()

    for proposal in proposals:
        allowed_extra_keys.update(set(signature(proposal).parameters.keys()))

    invalid_keys = extra_keys - allowed_extra_keys

    if invalid_keys:
        raise RuntimeError(
            f"Unknown kwargs for {ProposalClass.__name__}: {invalid_keys}."
        )
    else:
        logger.warning(
            f"Removing unused keyword arguments ({extra_keys}) from kwargs for"
            f" {ProposalClass.__name__}. These are valid keyword arguments but"
            " correspond to other proposal classes."
        )
        for key in extra_keys:
            kwargs_out.pop(key)
    return kwargs_out



"""Configuration for ImportanceFlowProposal tests"""

import pytest
from unittest.mock import MagicMock, create_autospec

from nessai.livepoint import numpy_array_to_live_points
from nessai.model import Model
from nessai.proposal.importance import ImportanceFlowProposal
import numpy as np

NSAMPLES = 10


@pytest.fixture()
def ifp():
    obj = create_autospec(ImportanceFlowProposal)
    obj.model = MagicMock(spec=Model)
    return obj


@pytest.fixture()
def x_array(model):
    return np.random.randn(NSAMPLES, model.dims)


@pytest.fixture
def x(x_array, model):
    return numpy_array_to_live_points(x_array, model.names)


@pytest.fixture()
def x_prime(model):
    return np.random.randn(NSAMPLES, model.dims)


@pytest.fixture
def log_j():
    return np.random.randn(NSAMPLES)



# -*- coding: utf-8 -*-
"""
Null reparameterisation.
"""
import logging

from .base import Reparameterisation

logger = logging.getLogger(__name__)


class NullReparameterisation(Reparameterisation):
    """Reparameterisation that does not modify the parameters.

    Parameters
    ----------
    parameters : list or str
        Parameters for which the reparameterisation will be used.
    prior_bounds : list, dict or None
        Prior bounds for parameters. Unused for this reparameterisation.
    """

    def __init__(self, parameters=None, prior_bounds=None):
        super().__init__(parameters=parameters, prior_bounds=prior_bounds)
        self.prime_parameters = self.parameters
        logger.debug(f"Initialised reparameterisation: {self.name}")

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : Log jacobian to be updated
        """
        x_prime[self.prime_parameters] = x[self.parameters]
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : Log jacobian to be updated
        """
        x[self.parameters] = x_prime[self.prime_parameters]
        return x, x_prime, log_j



# -*- coding: utf-8 -*-
"""Tests for the nessai.flows.nets submodule"""

from unittest.mock import create_autospec

import torch
from torch import nn
import torch.nn.functional as F
import pytest

from nessai.flows.nets import MLP


@pytest.fixture
def mlp():
    return create_autospec(MLP)


def test_mlp_init_defaults(mlp):
    """Assert the default init returns the correct network"""
    MLP.__init__(mlp, (2,), (1,), [64, 64])
    assert len(mlp._hidden_layers) == 1


@pytest.mark.parametrize("activate_output", [False, True, F.relu])
def test_mlp_init_activate_output(mlp, activate_output):
    """Assert the default init returns the correct network"""
    MLP.__init__(mlp, (2,), (1,), [64, 64], activate_output=activate_output)


def test_init_invalid_hidden(mlp):
    """Assert an error is raised if the hidden sizes list is empty"""
    with pytest.raises(ValueError) as excinfo:
        MLP.__init__(mlp, (2,), (1,), [])
    assert "List of hidden sizes" in str(excinfo.value)


def test_init_invalid_output_act(mlp):
    """Assert an error is raised activate output is the wrong type"""
    with pytest.raises(TypeError) as excinfo:
        MLP.__init__(mlp, (2,), (1,), [10], activate_output="relu")
    assert "must be a boolean or a callable" in str(excinfo.value)


def test_mlp_forward(mlp):
    """Test the MLP implementation of forward"""
    x = torch.rand(4, 2)
    # All of the components of the MLP
    mlp._in_shape = torch.Size((2,))
    mlp._out_shape = torch.Size((1,))
    mlp._input_layer = nn.Linear(2, 10)
    mlp._hidden_layers = [nn.Linear(10, 10)]
    mlp._dropout_layers = [nn.Dropout(0.5)]
    mlp._output_layer = nn.Linear(10, 1)
    mlp._activate_output = F.sigmoid
    mlp._activation = F.relu
    mlp._output_activation = F.sigmoid
    out = MLP.forward(mlp, x, context=None)
    assert out.shape == torch.Size((4, 1))
    assert all(out >= 0.0)
    assert all(out <= 1.0)


def test_mlp_forward_context(mlp):
    """Assert an error is raised in the context is not None"""
    x = torch.tensor(1)
    with pytest.raises(ValueError) as excinfo:
        MLP.forward(mlp, x, context=x)
    assert "MLP with conditional inputs is not implemented." in str(
        excinfo.value
    )


def test_mlp_forward_incorrect_shape(mlp):
    """Assert an error is raised if the input shape is incorrect"""
    x = torch.rand(10, 2)
    mlp._in_shape = torch.Size((3,))
    with pytest.raises(ValueError) as excinfo:
        MLP.forward(mlp, x)
    assert "Expected inputs of shape" in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Tests for rescaling functions
"""
import numpy as np
import pytest
from unittest.mock import patch


from nessai.utils.rescaling import (
    configure_edge_detection,
    detect_edge,
    determine_rescaled_bounds,
    exp_with_log_jacobian,
    inverse_rescale_minus_one_to_one,
    inverse_rescale_zero_to_one,
    logistic_function,
    log_with_log_jacobian,
    rescale_minus_one_to_one,
    rescale_zero_to_one,
    logit,
    sigmoid,
)


def test_rescale_minus_one_to_one():
    """Assert rescaling is correctly applied."""
    x = np.array([-5.0, -2.5, 0.0, 2.5, 5.0])
    expected = np.array([-1, -0.5, 0.0, 0.5, 1.0])
    x_out, log_j = rescale_minus_one_to_one(x, -5, 5)
    np.testing.assert_array_equal(x_out, expected)
    np.testing.assert_equal(log_j, np.log(2) - np.log(10))


def test_inverse_rescale_minus_one_to_one():
    """Assert rescaling is correctly applied."""
    expected = np.array([-5.0, -2.5, 0.0, 2.5, 5.0])
    x = np.array([-1, -0.5, 0.0, 0.5, 1.0])
    x_out, log_j = inverse_rescale_minus_one_to_one(x, -5, 5)
    np.testing.assert_array_equal(x_out, expected)
    np.testing.assert_equal(log_j, -np.log(2) + np.log(10))


def test_rescale_zero_to_one():
    """Assert rescaling is correctly applied."""
    x = np.array([-5.0, -2.5, 0.0, 2.5, 5.0])
    expected = np.array([0, 0.25, 0.5, 0.75, 1.0])
    x_out, log_j = rescale_zero_to_one(x, -5, 5)
    np.testing.assert_array_equal(x_out, expected)
    np.testing.assert_equal(log_j, -np.log(10))


def test_inverse_rescale_zero_to_one():
    """Assert rescaling is correctly applied."""
    expected = np.array([-5.0, -2.5, 0.0, 2.5, 5.0])
    x = np.array([0.0, 0.25, 0.5, 0.75, 1.0])
    x_out, log_j = inverse_rescale_zero_to_one(x, -5, 5)
    np.testing.assert_array_equal(x_out, expected)
    np.testing.assert_equal(log_j, np.log(10))


@pytest.mark.parametrize(
    "test, expected",
    [(False, False), ("lower", False), ("upper", "upper"), ("both", "both")],
)
def test_detect_edge_test(test, expected):
    """Test detect edge test mode which should skip the main function."""
    with patch("numpy.histogram") as m:
        out = detect_edge(1, test=test, allowed_bounds=["upper"])
    assert out == expected
    m.assert_not_called()


@pytest.mark.parametrize(
    "hist_values, kwargs, expected",
    [
        ([0, 0, 10], {}, "upper"),
        ([10, 0, 0], {}, "lower"),
        ([0, 10, 0], {"allow_none": True}, False),
        ([10, 0, 10], {"allow_both": True}, "both"),
        ([0.2, 10, 0.1], {"allow_none": False}, "lower"),
        ([10, 0, 0], {"allowed_bounds": ["upper"], "allow_none": True}, False),
    ],
)
def test_detect_edge_max_location(hist_values, kwargs, expected):
    """Test detect edge based on location of the max"""
    with patch("numpy.histogram", return_value=(hist_values, [1, 2, 3])) as m:
        out = detect_edge(
            [5, 6], nbins=3, x_range=[-2, 2], percent=0.1, **kwargs
        )
    m.assert_called_once_with([5, 6], bins=3, density=True, range=[-2, 2])
    assert out == expected


def test_detect_edge_auto_bins():
    """Assert auto bins is called if nbins is auto"""
    with patch(
        "numpy.histogram", return_value=([10, 0, 0], [1, 2, 3])
    ) as m, patch("nessai.utils.rescaling.auto_bins", return_value=4) as mab:
        out = detect_edge([5, 6], nbins="auto")
    m.assert_called_once_with([5, 6], bins=4, density=True, range=None)
    mab.assert_called_once_with([5, 6])
    assert out == "lower"


def test_detect_edge_invalid_bound():
    """Assert that invalid allowed bounds raise an error"""
    with pytest.raises(RuntimeError) as excinfo:
        detect_edge(1, allowed_bounds=["both"])
    assert "Unknown allowed bounds: ['both']" in str(excinfo.value)


def test_configure_edge_detection_detect_edges():
    """Test configuring edge detection."""
    expected = {"x": 1, "allow_none": True, "cutoff": 0.5}
    out = configure_edge_detection({"x": 1}, detect_edges=True)
    assert out == expected


def test_configure_edge_detection_no_detect_edges():
    """Test configuring edge detection when set to False.

    Also test behaviour when the dictionary is None.
    """
    expected = {"allow_none": False, "cutoff": 0.0}
    out = configure_edge_detection(None, False)
    assert out == expected


@pytest.mark.parametrize(
    "prior_min, prior_max, x_min, x_max, kwargs, expected",
    [
        (-10, 8, -2, 2, {"inversion": False}, (-5, 4)),
        (-10, 8, -2, 2, {"inversion": True, "invert": False}, (-5, 4)),
        (-10, 8, -2, 2, {"inversion": False, "invert": "lower"}, (-5, 4)),
        (-10, 6, -2, 2, {"invert": "lower", "inversion": True}, (-2, 2)),
        (-10, 6, -2, 2, {"invert": "upper", "inversion": True}, (-3, 3)),
        (-10, 6, -2, 2, {"invert": "both", "inversion": True}, (-0.5, 1.5)),
    ],
)
def test_determine_rescaled_bounds(
    prior_min, prior_max, x_min, x_max, kwargs, expected
):
    """Assert the correct rescaled prior bounds are returned."""
    out = determine_rescaled_bounds(
        prior_min, prior_max, x_min, x_max, **kwargs
    )
    assert out == expected


def test_determine_rescaled_bounds_min_max_equal():
    """Assert an error is raised if the min and max are equal;"""
    with pytest.raises(ValueError) as excinfo:
        determine_rescaled_bounds(-1, 1, 0.5, 0.5)
    assert "New minimum and maximum are equal" in str(excinfo.value)


def test_determine_rescaled_bounds_invalid_invert():
    """Assert an error is raised if invert is not a valid value."""
    with pytest.raises(ValueError) as excinfo:
        determine_rescaled_bounds(
            -1, 1, -0.5, 0.5, invert="test", inversion=True
        )
    assert "Invalid value for `invert`: test" in str(excinfo.value)


@pytest.mark.parametrize(
    "x, y, log_J", [(0.0, -np.inf, np.inf), (1.0, np.inf, np.inf)]
)
def test_logit_bounds(x, y, log_J):
    """
    Test logit at the bounds
    """
    with pytest.warns(RuntimeWarning):
        assert logit(x, eps=0) == (y, log_J)


@pytest.mark.parametrize(
    "x, y, log_J", [(np.inf, 1, -np.inf), (-np.inf, 0, -np.inf)]
)
def test_sigmoid_bounds(x, y, log_J):
    """
    Test sigmoid for inf
    """
    assert sigmoid(x) == (y, log_J)


@pytest.mark.parametrize("p", [1e-5, 0.5, 1.0 - 1e-5])
@pytest.mark.parametrize("eps", [False, 1e-12])
def test_logit_sigmoid(p, eps):
    """
    Test invertibility of sigmoid(logit(x))
    """
    x = logit(p, eps=eps)
    y = sigmoid(x[0])
    np.testing.assert_almost_equal(p, y[0], decimal=10)
    np.testing.assert_almost_equal(x[1] + y[1], 0.0, decimal=10)


@pytest.mark.parametrize("p", [-10.0, -1.0, 0.0, 1.0, 10.0])
@pytest.mark.parametrize("eps", [False, 1e-12])
def test_sigmoid_logit(p, eps):
    """
    Test invertibility of logit(sigmoid(x))
    """
    x = sigmoid(p)
    y = logit(x[0], eps=eps)
    np.testing.assert_almost_equal(p, y[0], decimal=10)
    np.testing.assert_almost_equal(x[1] + y[1], 0.0, decimal=10)


def test_logistic_function():
    """Assert correct value is returned when k and x0 are specified"""
    assert logistic_function(0.0, 3.0, 2.0) == (1 / (1 + np.exp(6)))


def test_logistic_function_reference():
    """Compare the logistic function to a reference from scipy"""
    from scipy.special import expit

    x = np.array([-5.0, 0.0, 5.0])
    np.testing.assert_array_equal(logistic_function(x), expit(x))


def test_log():
    """Assert the correct values are returned"""
    x = np.random.rand(10)
    x_log, logj_log = log_with_log_jacobian(x)
    np.testing.assert_array_equal(x_log, np.log(x))
    np.testing.assert_array_equal(logj_log, -np.log(x))


def test_exp():
    """Assert the correct values are returned"""
    x = np.random.randn(10)
    x_exp, logj = exp_with_log_jacobian(x)
    np.testing.assert_array_equal(x_exp, np.exp(x))
    np.testing.assert_array_equal(logj, x)


def test_log_exp_inverse():
    """Assert the log and exp functions are the inverse of each other"""
    x = np.random.rand(10)
    x_log, logj_log = log_with_log_jacobian(x)
    x_out, logj_exp = exp_with_log_jacobian(x_log)
    np.testing.assert_almost_equal(x_out, x, decimal=14)
    np.testing.assert_almost_equal(logj_log, -logj_exp, decimal=14)



# -*- coding: utf-8 -*-
"""
Test functions related to training and using the flow.
"""
import os
import numpy as np
import pytest
from unittest.mock import MagicMock, patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.proposal import FlowProposal
from nessai.utils.testing import assert_structured_arrays_equal


def test_reset_model_weights(proposal):
    """Test resetting model weights"""
    proposal.flow = MagicMock()
    proposal.flow.reset_model = MagicMock()
    FlowProposal.reset_model_weights(proposal, reset_permutations=True)
    proposal.flow.reset_model.assert_called_once_with(reset_permutations=True)


def test_train_not_initialised(proposal):
    """Assert an error is raised if the proposal is not initialised"""
    proposal.initialised = False
    with pytest.raises(
        RuntimeError, match=r"FlowProposal is not initialised."
    ):
        FlowProposal.train(proposal, [1, 2])


@patch("os.path.exists", return_value=False)
@patch("os.makedirs")
def test_train_plot_false(mock_os_makedirs, proposal, model):
    """Test the train method"""
    x = model.new_point(2)
    x_prime = model.new_point(2)
    proposal.rescaled_names = model.names
    proposal.save_training_data = False
    proposal.training_count = 0
    proposal.populated = True
    proposal.flow = MagicMock()
    proposal.flow.train = MagicMock()
    proposal.check_state = MagicMock()
    proposal.rescale = MagicMock(return_value=(x_prime, np.zeros_like(x)))
    FlowProposal.train(proposal, x, plot=False)

    assert_structured_arrays_equal(x, proposal.training_data)
    proposal.check_state.assert_called_once_with(proposal.training_data)
    proposal.rescale.assert_called_once_with(x)
    assert proposal.populated is False
    assert proposal.training_count == 1
    proposal.flow.train.assert_called_once()


@pytest.mark.parametrize("n", [1, 10])
def test_forward_pass(proposal, model, n):
    """Test the forward pass method"""
    x = model.new_point(n)
    z = np.random.randn(n, model.dims)
    proposal.clip = False
    proposal.rescale = MagicMock(return_value=[x, 2 * np.ones(n)])
    proposal.rescaled_names = model.names
    proposal.flow = MagicMock()
    proposal.flow.forward_and_log_prob = MagicMock(
        return_value=[z, np.ones(n)]
    )

    z_out, log_p = FlowProposal.forward_pass(proposal, x, compute_radius=False)

    assert np.array_equal(z, z_out)
    assert np.array_equal(log_p, 3 * np.ones(n))
    proposal.rescale.assert_called_once_with(x, compute_radius=False)
    proposal.flow.forward_and_log_prob.assert_called_once()


@pytest.mark.parametrize("log_p", [np.ones(2), np.array([-1, np.inf])])
@pytest.mark.parametrize("discard_nans", [False, True])
def test_backward_pass(proposal, model, log_p, discard_nans):
    """Test the forward pass method"""
    n = 2
    if discard_nans:
        acc = int(np.isfinite(log_p).sum())
    else:
        acc = len(log_p)
    x = np.random.randn(n, model.dims)
    z = np.random.randn(n, model.dims)
    proposal.inverse_rescale = MagicMock(
        side_effect=lambda a: (a, np.ones(a.size))
    )
    proposal.rescaled_names = model.names
    proposal.alt_dist = None
    proposal.check_prior_bounds = MagicMock(
        side_effect=lambda a, b, c: (a, b, c)
    )
    proposal.flow = MagicMock()
    proposal.flow.sample_and_log_prob = MagicMock(return_value=[x, log_p])

    x_out, log_p = FlowProposal.backward_pass(
        proposal,
        z,
        discard_nans=discard_nans,
    )

    assert len(x_out) == acc
    proposal.inverse_rescale.assert_called_once()
    proposal.flow.sample_and_log_prob.assert_called_once_with(
        z=z, alt_dist=None
    )


@pytest.mark.parametrize("save", [True, False])
@pytest.mark.parametrize("plot", [True, False])
@pytest.mark.parametrize("plot_training", [True, False])
def test_training(proposal, tmpdir, save, plot, plot_training):
    """Test the training method"""
    output = tmpdir.mkdir("test")
    data = np.random.randn(10, 2)
    data_prime = data / 2
    x = numpy_array_to_live_points(data, ["x", "y"])
    x_prime = numpy_array_to_live_points(data_prime, ["x_prime", "y_prime"])
    log_j = np.ones(data.shape[0])

    proposal.initialised = True
    proposal.training_count = 0
    proposal.populated = True
    proposal._plot_training = plot_training
    proposal.save_training_data = save
    proposal.rescaled_names = ["x_prime", "y_prime"]
    proposal.output = output

    proposal.check_state = MagicMock()
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.flow = MagicMock()
    proposal.flow.train = MagicMock()
    proposal._plot_training_data = MagicMock()

    with patch(
        "nessai.proposal.flowproposal.live_points_to_array",
        return_value=data_prime,
    ), patch("nessai.proposal.flowproposal.save_live_points") as mock_save:
        FlowProposal.train(proposal, x, plot=plot)

    assert_structured_arrays_equal(x, proposal.training_data)

    if save or (plot and plot_training):
        output = os.path.join(output, "training", "block_0", "")

    if save:
        mock_save.assert_called_once()

    if plot and plot_training:
        proposal._plot_training_data.assert_called_once_with(output)
    elif not plot or not plot_training:
        proposal._plot_training_data.assert_not_called()

    proposal.check_state.assert_called_once_with(proposal.training_data)
    proposal.rescale.assert_called_once_with(x)
    proposal.flow.train.assert_called_once_with(
        data_prime, output=output, plot=plot and plot_training
    )
    assert proposal.training_count == 1
    assert proposal.populated is False



# -*- coding: utf-8 -*-
"""
Test the FlowModel utils.
"""
import pytest
from unittest.mock import patch

from nessai.flowmodel.config import DEFAULT_MODEL_CONFIG
from nessai.flowmodel.utils import update_config, update_model_config


def test_update_config():
    """Test with a basic config"""
    model_config = dict(n_blocks=2)
    config = dict(patience=1, model_config=model_config)
    with patch(
        "nessai.flowmodel.utils.update_model_config",
        return_value=DEFAULT_MODEL_CONFIG.copy(),
    ) as mock_update:
        out = update_config(config)
    mock_update.assert_called_once_with(model_config)
    assert out["patience"] == 1


def test_update_config_none():
    """Test update config for no input"""
    with patch(
        "nessai.flowmodel.utils.update_model_config",
        return_value=DEFAULT_MODEL_CONFIG.copy(),
    ) as mock_update:
        config = update_config(None)
    mock_update.assert_called_once_with(None)
    assert "model_config" in config


def test_update_config_invalid_type():
    """Test update config when an invalid argument is specified"""
    with pytest.raises(TypeError) as excinfo:
        update_config(False)
    assert "Must pass a dictionary" in str(excinfo.value)


def test_update_config_invalid_noise_scale():
    """Assert an error is raised if noise_scale is not a float"""
    config = {"noise_scale": "auto", "noise_type": "adaptive"}
    with pytest.raises(TypeError) as excinfo:
        update_config(config)
    assert "`noise_scale` must be a float" in str(excinfo.value)


def test_update_config_missing_noise_type():
    """Assert noise type is set if only `noise_scale` is given"""
    config = {"noise_scale": 1.0}
    out = update_config(config)
    assert out["noise_type"] == "constant"


def test_update_config_missing_noise_scale():
    """Assert an error is raised if noise_scaling is missing when noise_type is
    set.
    """
    config = {"noise_type": "constant"}
    with pytest.raises(RuntimeError) as excinfo:
        update_config(config)
    assert "`noise_scale` must be specified" in str(excinfo.value)


def test_update_model_config_n_neurons():
    """Assert `get_n_neurons` is called"""
    config = dict(n_inputs=10)
    with patch(
        "nessai.flowmodel.utils.get_n_neurons", return_value=20
    ) as mock:
        config = update_model_config(config)
    mock.assert_called_once_with(n_neurons=None, n_inputs=10)
    assert config["n_neurons"] == 20


def test_update_model_config_none():
    """Assert the default is returned if the inputs is None"""
    assert update_model_config(None) == DEFAULT_MODEL_CONFIG


def test_update_model_config_type_error():
    """Assert a type error is raised if the input is the wrong type"""
    with pytest.raises(TypeError) as excinfo:
        update_model_config(False)
    assert "Must pass a dictionary" in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Reparameterisations for handling angles.
"""
import logging

import numpy as np
from scipy import stats

from .base import Reparameterisation
from ..priors import (
    log_2d_cartesian_prior,
    log_2d_cartesian_prior_sine,
)
from ..utils.rescaling import rescale_zero_to_one, inverse_rescale_zero_to_one

logger = logging.getLogger(__name__)


class Angle(Reparameterisation):
    """Reparameterisation for a single angle.

    This reparameterisations converts an angle to Cartesian coordinates
    using either a corresponding radial parameter or an auxiliary radial
    parameter. When using the auxiliary parameter, samples are drawn from
    a chi-distribution with two degrees of freedom.

    Parameters
    ----------
    parameters : str or list
        Parameter(s) to use for the reparameterisation. Either just an angle
        or an angle and corresponding radial parameter
    prior_bounds : dict
        Dictionary of prior bounds. Does not need to be specified when defining
        `reparameterisations`.
    scale : float, optional
        Value used to rescale the angle before converting to Cartesian
        coordinates. If None the scale will be set to 2pi / prior_bounds.
    prior : {'uniform', 'sine', None}
        Type of prior being used for sampling this angle. If specified, the
        prime prior is enabled. If None then it is disabled.
    """

    requires_bounded_prior = True

    def __init__(
        self, parameters=None, prior_bounds=None, scale=1.0, prior=None
    ):
        super().__init__(parameters=parameters, prior_bounds=prior_bounds)

        if len(self.parameters) == 1:
            self.parameters.append(self.parameters[0] + "_radial")
            self.chi = stats.chi(2)
            self.has_prior = True
        elif len(self.parameters) == 2:
            self.chi = False
        else:
            raise RuntimeError("Too many parameters for Angle")

        if scale is None:
            logger.debug("Scale is None, using 2pi / prior_range")
            self.scale = 2.0 * np.pi / np.ptp(self.prior_bounds[self.angle])
        else:
            self.scale = scale

        if prior_bounds[self.angle][0] == 0:
            self._zero_bound = True
        else:
            self._zero_bound = False

        self.prime_parameters = [self.angle + "_x", self.angle + "_y"]
        self.requires = []

        if prior in ["uniform", "sine"]:
            self.prior = prior
            self.has_prime_prior = True
            if self.prior == "uniform":
                self._prime_prior = log_2d_cartesian_prior
                self._k = self.scale * np.pi
            else:
                self._prime_prior = log_2d_cartesian_prior_sine
                self._k = np.pi
            logger.debug(f"Prime prior enabled for {self.name}")
        else:
            self.has_prime_prior = False
            logger.debug(f"Prime prior disabled for {self.name}")

    @property
    def angle(self):
        """The name of the angular parameter"""
        return self.parameters[0]

    @property
    def radial(self):
        """The name of the radial parameter"""
        return self.parameters[1]

    @property
    def radius(self):
        """The name of the radial parameter (equivalent to radial)"""
        return self.parameters[1]

    @property
    def x(self):
        """The name of x coordinate"""
        return self.prime_parameters[0]

    @property
    def y(self):
        """The name of y coordinate"""
        return self.prime_parameters[1]

    def _rescale_radial(self, x, x_prime, log_j, **kwargs):
        return x[self.parameters[1]], x, x_prime, log_j

    def _rescale_angle(self, x, x_prime, log_j, **kwargs):
        return x[self.parameters[0]] * self.scale, x, x_prime, log_j

    def _inverse_rescale_angle(self, x, x_prime, log_j):
        return x, x_prime, log_j

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """Convert the angle to Cartesian coordinates"""
        angle, x, x_prime, log_j = self._rescale_angle(
            x, x_prime, log_j, **kwargs
        )

        if self.chi:
            r = self.chi.rvs(size=x.size)
        else:
            r, x, x_prime, log_j = self._rescale_radial(
                x, x_prime, log_j, **kwargs
            )
        if any(r < 0):
            raise RuntimeError("Radius cannot be negative.")

        x_prime[self.prime_parameters[0]] = r * np.cos(angle)
        x_prime[self.prime_parameters[1]] = r * np.sin(angle)
        log_j += np.log(r)
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """Convert from Cartesian to an angle and radial component"""
        x[self.parameters[1]] = np.sqrt(
            x_prime[self.prime_parameters[0]] ** 2
            + x_prime[self.prime_parameters[1]] ** 2
        )
        if self._zero_bound:
            x[self.parameters[0]] = (
                np.arctan2(
                    x_prime[self.prime_parameters[1]],
                    x_prime[self.prime_parameters[0]],
                )
                % (2.0 * np.pi)
                / self.scale
            )
        else:
            x[self.parameters[0]] = (
                np.arctan2(
                    x_prime[self.prime_parameters[1]],
                    x_prime[self.prime_parameters[0]],
                )
                / self.scale
            )

        log_j -= np.log(x[self.parameters[1]])
        x, x_prime, log_j = self._inverse_rescale_angle(x, x_prime, log_j)

        return x, x_prime, log_j

    def log_prior(self, x):
        """Prior for radial parameter"""
        return self.chi.logpdf(x[self.parameters[1]])

    def x_prime_log_prior(self, x_prime):
        """Compute the prior in the prime space assuming a uniform prior"""
        if self.has_prime_prior:
            return self._prime_prior(
                x_prime[self.prime_parameters[0]],
                x_prime[self.prime_parameters[1]],
                k=self._k,
            )
        else:
            raise RuntimeError("Prime prior")


class ToCartesian(Angle):
    """Convert a parameter to Cartesian coordinates"""

    def __init__(self, mode="split", scale=np.pi, **kwargs):
        super().__init__(scale=scale, **kwargs)

        self.mode = mode
        if self.mode not in ["duplicate", "split", "half"]:
            raise RuntimeError(f"Unknown mode: {self.mode}")
        logger.debug(f"Using mode: {self.mode}")

        self._zero_bound = False
        self._k = self.prior_bounds[self.parameters[0]][1]

    def _rescale_angle(
        self, x, x_prime, log_j, compute_radius=False, **kwargs
    ):
        angle, lj = rescale_zero_to_one(
            x[self.parameters[0]], *self.prior_bounds[self.parameters[0]]
        )
        log_j += lj
        if self.mode == "duplicate" or compute_radius:
            angle = np.concatenate([angle, -angle])
            x = np.concatenate([x, x])
            x_prime = np.concatenate([x_prime, x_prime])
            log_j = np.concatenate([log_j, log_j])
        elif self.mode == "split":
            neg = np.random.choice(angle.size, angle.size // 2, replace=False)
            angle[neg] *= -1

        angle *= self.scale
        return angle, x, x_prime, log_j

    def _inverse_rescale_angle(self, x, x_prime, log_j):
        x[self.parameters[0]], lj = inverse_rescale_zero_to_one(
            np.abs(x[self.parameters[0]]),
            *self.prior_bounds[self.parameters[0]],
        )
        log_j += lj
        return x, x_prime, log_j


class AnglePair(Reparameterisation):
    """Reparameterisation for a pair of angles and a radial component.

    Converts to three-dimensional Cartesian coordinates.

    If the radial component is not specified, it is sampled from a chi-
    distribution with three degrees of freedom.

    Notes
    -----
    The parameters will be reordered such that the first parameter is the angle
    along the horizon, the second parameter is the vertical angle and the last
    parameter is the radial parameter.

    Parameters
    ----------
    parameters : list
        List of parameters. Must contain at least the two angles and,
        optionally, also a radial component.
    prior_bounds : dict
        Dictionary of prior bounds for each parameter
    prior : str, {'isotropic', None}
        Type of prior, used to enable use of the prime prior.
    convention : str, {'ra-dec', 'az-zen'}
        Convention used for defining the spherical polar coordinates. If not
        set, it will be guessed based on either dec or zen. Where it is assumed
        declination is defined on [-pi/2, pi/2] and zenith on [0, pi].
    """

    requires_bounded_prior = True
    known_priors = ["isotropic", None]
    _conventions = {"az-zen": [0, np.pi], "ra-dec": [-np.pi / 2, np.pi / 2]}

    def __init__(
        self, parameters=None, prior_bounds=None, prior=None, convention=None
    ):

        if len(parameters) not in [2, 3]:
            raise RuntimeError(
                "Must use a pair of angles or a pair plus a radius"
            )

        super().__init__(parameters=parameters, prior_bounds=prior_bounds)

        # Determine which parameter is for the horizontal plane
        # and which is for the vertical plane
        logger.debug("Checking order of parameters")
        b = np.ptp([prior_bounds[p] for p in parameters], axis=1)
        try:
            hz = np.where(b == (2 * np.pi))[0][0]
            vt = np.where(b == np.pi)[0][0]
        except IndexError:
            raise ValueError(
                f"Invalid prior ranges: {self.prior_bounds}. "
                "Parameters must be defined over a range of pi and 2 pi!"
            )
        # Make sure order is horizon, vertical, radial
        if len(parameters) == 3:
            r = list({0, 1, 2} - {hz, vt})[0]
            parameters[0], parameters[1], parameters[2] = (
                parameters[hz],
                parameters[vt],
                parameters[r],
            )
            m = "_".join(parameters)
            self.chi = False
        else:
            parameters[0], parameters[1] = parameters[hz], parameters[vt]
            m = "_".join(parameters)
            parameters.append(f"{m}_radial")
            self.chi = stats.chi(3)
            self.has_prior = True

        logger.debug(f"Parameters are: {parameters}")

        # Modulo 2pi is used if the first parameter (ra/az) in defined on
        # [0, 2pi] since output would otherwise be [-pi, pi]
        if np.array_equal(self.prior_bounds[parameters[0]], [0, 2 * np.pi]):
            self._modulo_2pi = True
        elif np.array_equal(self.prior_bounds[parameters[0]], [-np.pi, np.pi]):
            self._modulo_2pi = False
        else:
            raise ValueError(
                f"Prior bounds for {parameters[0]} must be [0, 2pi] or "
                f"[-pi, pi]. Received: {self.prior_bounds[parameters[0]]}"
            )

        self.parameters = parameters
        self.prime_parameters = [f"{m}_{x}" for x in ["x", "y", "z"]]

        if prior == "isotropic" and self.chi:
            self.has_prime_prior = True
            logger.info(f"Prime prior enabled for {self.name}")
        elif prior not in self.known_priors:
            raise ValueError(
                f"Unknown prior: `{prior}`. Choose from: {self.known_priors}"
            )
        else:
            self.has_prime_prior = False
            logger.debug(f"Prime prior disabled for {self.name}")

        if convention is None:
            logger.debug("Trying to determine convention")
            if np.array_equal(
                self.prior_bounds[self.parameters[1]], [0, np.pi]
            ):
                self.convention = "az-zen"
            elif np.array_equal(
                self.prior_bounds[self.parameters[1]], [-np.pi / 2, np.pi / 2]
            ):
                self.convention = "ra-dec"
            else:
                raise RuntimeError(
                    f"Could not determine convention for: {self.parameters}!"
                )
        elif convention in self._conventions.keys():
            self.convention = convention
            if not np.array_equal(
                self.prior_bounds[parameters[1]], self._conventions[convention]
            ):
                raise ValueError(
                    f"Prior bounds for {parameters[1]} must be "
                    f"{self._conventions[convention]} for the "
                    f"{convention} convention. "
                    f"Received: {self.prior_bounds[parameters[1]]}."
                )
        else:
            raise ValueError(
                f"Unknown convention: `{convention}`. "
                f"Choose from: {list(self._conventions.keys())}."
            )

        logger.debug(f"Using convention: {self.convention}")

        self.requires = []

    @property
    def angles(self):
        """Names of the two angles.

        Order is: angle along the horizon, vertical angle.
        """
        return self.parameters[:2]

    @property
    def radial(self):
        """Name of the radial parameter"""
        return self.parameters[-1]

    @property
    def x(self):
        """Name of the first Cartesian coordinate"""
        return self.prime_parameters[0]

    @property
    def y(self):
        """Name of the second Cartesian coordinate"""
        return self.prime_parameters[1]

    @property
    def z(self):
        """Name of the third Cartesian coordinate"""
        return self.prime_parameters[2]

    def _az_zen(self, x, x_prime, log_j, r):
        x_prime[self.prime_parameters[0]] = (
            r * np.sin(x[self.parameters[1]]) * np.cos(x[self.parameters[0]])
        )
        x_prime[self.prime_parameters[1]] = (
            r * np.sin(x[self.parameters[1]]) * np.sin(x[self.parameters[0]])
        )
        x_prime[self.prime_parameters[2]] = r * np.cos(x[self.parameters[1]])
        log_j += 2 * np.log(r) + np.log(np.sin(x[self.parameters[1]]))
        return x, x_prime, log_j

    def _ra_dec(self, x, x_prime, log_j, r):
        x_prime[self.prime_parameters[0]] = (
            r * np.cos(x[self.parameters[1]]) * np.cos(x[self.parameters[0]])
        )
        x_prime[self.prime_parameters[1]] = (
            r * np.cos(x[self.parameters[1]]) * np.sin(x[self.parameters[0]])
        )
        x_prime[self.prime_parameters[2]] = r * np.sin(x[self.parameters[1]])
        log_j += 2 * np.log(r) + np.log(np.cos(x[self.parameters[1]]))
        return x, x_prime, log_j

    def _inv_az_zen(self, x, x_prime, log_j):
        x[self.parameters[2]] = np.sqrt(
            x_prime[self.prime_parameters[0]] ** 2.0
            + x_prime[self.prime_parameters[1]] ** 2.0
            + x_prime[self.prime_parameters[2]] ** 2.0
        )

        if self._modulo_2pi:
            x[self.parameters[0]] = np.arctan2(
                x_prime[self.prime_parameters[1]],
                x_prime[self.prime_parameters[0]],
            ) % (2.0 * np.pi)
        else:
            x[self.parameters[0]] = np.arctan2(
                x_prime[self.prime_parameters[1]],
                x_prime[self.prime_parameters[0]],
            )
        x[self.parameters[1]] = np.arctan2(
            np.sqrt(
                x_prime[self.prime_parameters[0]] ** 2.0
                + x_prime[self.prime_parameters[1]] ** 2.0
            ),
            x_prime[self.prime_parameters[2]],
        )
        log_j += -2 * np.log(x[self.parameters[2]]) - np.log(
            np.sin(x[self.parameters[1]])
        )

        return x, x_prime, log_j

    def _inv_ra_dec(self, x, x_prime, log_j):
        x[self.parameters[2]] = np.sqrt(
            x_prime[self.prime_parameters[0]] ** 2.0
            + x_prime[self.prime_parameters[1]] ** 2.0
            + x_prime[self.prime_parameters[2]] ** 2.0
        )

        if self._modulo_2pi:
            x[self.parameters[0]] = np.arctan2(
                x_prime[self.prime_parameters[1]],
                x_prime[self.prime_parameters[0]],
            ) % (2.0 * np.pi)
        else:
            x[self.parameters[0]] = np.arctan2(
                x_prime[self.prime_parameters[1]],
                x_prime[self.prime_parameters[0]],
            )

        x[self.parameters[1]] = np.arctan2(
            x_prime[self.prime_parameters[2]],
            np.sqrt(
                x_prime[self.prime_parameters[0]] ** 2.0
                + x_prime[self.prime_parameters[1]] ** 2.0
            ),
        )

        log_j += -2 * np.log(x[self.parameters[2]]) - np.log(
            np.cos(x[self.parameters[1]])
        )
        return x, x_prime, log_j

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Convert the spherical polar angles to Cartesian coordinates
        """
        if self.chi:
            r = self.chi.rvs(size=x.size)
        else:
            r = x[self.radial]
        if any(r < 0):
            raise RuntimeError("Radius cannot be negative.")

        if self.convention == "az-zen":
            return self._az_zen(x, x_prime, log_j, r)
        else:
            return self._ra_dec(x, x_prime, log_j, r)

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """Convert from Cartesian to spherical polar angles"""
        if self.convention == "az-zen":
            return self._inv_az_zen(x, x_prime, log_j)
        else:
            return self._inv_ra_dec(x, x_prime, log_j)

    def log_prior(self, x):
        """Prior for radial parameter"""
        if self.chi and self.has_prior:
            return self.chi.logpdf(x[self.parameters[2]])
        else:
            raise RuntimeError(
                "log_prior is not defined when a radial parameter has been "
                "specified!"
            )

    def x_prime_log_prior(self, x_prime):
        """
        Log probability of 3d Cartesian coordinates for an isotropic
        distribution of angles and a radial component drawn from a chi
        distribution with three degrees of freedom.
        """
        if self.has_prime_prior:
            return (
                -1.5 * np.log(2 * np.pi)
                - np.sum(
                    [x_prime[pp] ** 2 for pp in self.prime_parameters], axis=0
                )
                / 2
            )
        else:
            raise RuntimeError("x prime prior is not defined!")



"""Test the general configuration for ImportanceFlowProposal"""

import pytest
from unittest.mock import MagicMock, patch

from nessai.livepoint import (
    add_extra_parameters_to_live_points,
    reset_extra_live_points_parameters,
)
from nessai.proposal.importance import ImportanceFlowProposal as IFP


def test_init(ifp, model, tmp_path):
    """Test the init method"""
    output = tmp_path / "test"
    initial_draws = 1000
    IFP.__init__(ifp, model, output, initial_draws)
    assert ifp._weights[-1] == 1


@pytest.mark.usefixtures("ins_parameters")
def test_check_fields_pass():
    """Assert the checks pass"""
    IFP._check_fields()


def test_check_fields_logQ():
    """Assert an error is raised if logQ is missing"""
    with pytest.raises(
        RuntimeError, match=r"logQ field missing in non-sampling parameters."
    ):
        IFP._check_fields()


def test_check_fields_logW():
    """Assert an error is raised if logW is missing"""
    add_extra_parameters_to_live_points(["logQ"])
    with pytest.raises(
        RuntimeError, match=r"logW field missing in non-sampling parameters."
    ):
        IFP._check_fields()
    reset_extra_live_points_parameters()


def test_initialise(ifp):
    output = "test_init"
    flow_config = {"a": 1}
    ifp.initialised = False
    ifp.output = output
    ifp.flow_config = flow_config
    flow = MagicMock()
    with patch(
        "nessai.proposal.importance.Proposal.initialise",
    ) as mock_init, patch(
        "nessai.proposal.importance.ImportanceFlowModel", return_value=flow
    ) as mock_flow:
        IFP.initialise(ifp)
    ifp._check_fields.assert_called_once()
    ifp.verify_rescaling.assert_called_once()
    # Make sure flow is initialised
    mock_flow.assert_called_once_with(config=flow_config, output=output)
    flow.initialise.assert_called_once()
    mock_init.assert_called_once()


def test_already_initialised(ifp):
    """Assert functions are not called"""
    ifp.initialised = True

    with patch(
        "nessai.proposal.importance.Proposal.initialise",
    ) as mock_init, patch(
        "nessai.proposal.importance.ImportanceFlowModel",
    ) as mock_flow:
        IFP.initialise(ifp)
    mock_init.assert_not_called()
    mock_flow.assert_not_called()
    ifp._check_fields.assert_called_once()
    ifp.verify_rescaling.assert_not_called()



# -*- coding: utf-8 -*-
"""
Tests related to resuming.
"""
import os
import pytest
from unittest.mock import patch, MagicMock

from nessai.samplers.nestedsampler import NestedSampler


@pytest.fixture
def complete_sampler(model, tmpdir):
    """Complete instance of NestedSampler"""
    output = tmpdir.mkdir("output")
    ns = NestedSampler(model, output=output, poolsize=10)
    ns.initialise()
    return ns


def test_check_resume(sampler):
    """Test check resume method"""
    sampler.uninformed_sampling = False
    sampler.check_proposal_switch = MagicMock()
    sampler.resumed = True
    sampler._flow_proposal = MagicMock()
    sampler._flow_proposal.populated = False
    sampler._flow_proposal._resume_populated = True
    sampler._flow_proposal.indices = [1, 2, 3]

    NestedSampler.check_resume(sampler)

    sampler.check_proposal_switch.assert_called_once_with(force=True)
    assert sampler.resumed is False
    assert sampler._flow_proposal.populated is True


def test_check_resume_no_indices(sampler):
    """Test check resume method"""
    sampler.uninformed_sampling = True
    sampler.resumed = True
    sampler._flow_proposal = MagicMock()
    sampler._flow_proposal.populated = False
    sampler._flow_proposal._resume_populated = True
    sampler._flow_proposal.indices = []

    NestedSampler.check_resume(sampler)

    assert sampler.resumed is False
    assert sampler._flow_proposal.populated is False


@pytest.mark.parametrize("flow_config", [None, dict(a=1)])
def test_resume(model, flow_config):
    """Test the resume method"""
    obj = MagicMock()
    obj.model = model
    obj._uninformed_proposal = MagicMock()
    obj._uninformed_proposal.resume = MagicMock()
    obj._flow_proposal = MagicMock()
    obj._flow_proposal.resume = MagicMock()

    weights_file = "weight.pt"

    sampler = MagicMock()

    with patch(
        "nessai.samplers.base.BaseNestedSampler.resume_from_pickled_sampler",
        return_value=obj,
    ) as mock:
        out = NestedSampler.resume_from_pickled_sampler(
            sampler,
            model,
            flow_config=flow_config,
            weights_path=weights_file,
        )
    assert out is obj
    mock.assert_called_once_with(sampler, model)
    obj._uninformed_proposal.resume.assert_called_once_with(
        model,
    )
    obj._flow_proposal.resume.assert_called_once_with(
        model,
        flow_config if flow_config else {},
        weights_file,
    )


def test_get_state(sampler):
    """Test the getstate method used for pickling.

    It should remove the model.
    """
    sampler.model = MagicMock()
    state = NestedSampler.__getstate__(sampler)
    assert "model" not in state


@pytest.mark.integration_test
def test_checkpoint_integration(complete_sampler):
    """Integration test for checkpointing the sampler."""
    complete_sampler.checkpoint()
    resume_file = os.path.join(
        complete_sampler.output, complete_sampler.resume_file
    )
    assert os.path.exists(resume_file)


@pytest.mark.integration_test
def test_checkpoint_resume_integration(complete_sampler, model):
    """Integration test for checkpointing the sampler."""
    complete_sampler.likelihood_evaluations = [1, 2]
    complete_sampler.checkpoint()
    resume_file = os.path.join(
        complete_sampler.output, complete_sampler.resume_file
    )
    assert os.path.exists(resume_file)
    ns = NestedSampler.resume(resume_file, model)
    assert ns is not None



# -*- coding: utf-8 -*-
"""
Utilities for computing distances.
"""
import numpy as np
from scipy.spatial import distance


def compute_minimum_distances(samples, metric="euclidean"):
    """
    Compute the distance to the nearest neighbour of each sample

    Parameters
    ----------
    samples : array_like
        Array of samples.
    metric : str, optional
        Metric to use. See scipy docs for list of metrics:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.cdist.html

    Returns
    -------
    array_like
        Distance to nearest neighbour for each sample.
    """
    d = distance.cdist(samples, samples, metric)
    d[d == 0] = np.nan
    dmin = np.nanmin(d, axis=1)
    return dmin



# -*- coding: utf-8 -*-
"""
Proposals specifically for use with the importance based nested sampler.
"""
import logging
import os
from typing import Callable, Optional, Tuple, Union

import numpy as np
from scipy.special import logsumexp

from nessai.plot import plot_1d_comparison, plot_histogram, plot_live_points
from nessai.utils.testing import assert_structured_arrays_equal

from .base import Proposal
from .. import config
from ..flowmodel.importance import ImportanceFlowModel
from ..flowmodel.utils import update_config
from ..livepoint import (
    get_dtype,
    live_points_to_array,
    numpy_array_to_live_points,
)
from ..model import Model
from ..utils.rescaling import (
    logit,
    sigmoid,
)
from ..utils.structures import get_subset_arrays


logger = logging.getLogger(__name__)


class ImportanceFlowProposal(Proposal):
    """Flow-based proposal for importance-based nested sampling.

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User-defined model.
    output : str
        Output directory.
    flow_config : dict
        Configuration for the flow.
    reparameterisation : str
        Reparameterisation to use. If None, only the unit-hypercube
        reparameterisation is used.
    weighted_kl : bool
        If true, use the weights (prior / meta-proposal) when training the
        next flow.
    rest_flow : Union[bool, int]
        When to reset the flow. If False, the flow is not reset. If True the
        flow is reset every time a new flow is added. If an integer, every
        nth flow is reset.
    clip : bool
        If true the samples generated by flow will be clipped to [0, 1] before
        being mapped back from the unit-hypercube. This is only needed when
        the mapping cannot be defined outside of [0, 1]. In cases where it
        can, these points will be rejected when the prior bounds are checked.
    plot_training : bool
        If True, produce plots during training. If False, no plots are
        produced.
    """

    def __init__(
        self,
        model: Model,
        output: str,
        flow_config: dict = None,
        reparameterisation: str = "logit",
        weighted_kl: bool = True,
        reset_flow: Union[bool, int] = True,
        clip: bool = False,
        plot_training: bool = False,
    ) -> None:
        self.level_count = -1
        self._initialised = False

        self.model = model
        self.output = output
        self.flow_config = flow_config
        self.plot_training = plot_training
        self.reset_flow = int(reset_flow)
        self.reparameterisation = reparameterisation
        self.weighted_kl = weighted_kl
        self.clip = clip
        self._weights = {-1: 1.0}

        self.dtype = get_dtype(self.model.names)

    @property
    def weights(self) -> dict:
        """Dictionary containing the weights for each proposal"""
        return self._weights

    @property
    def weights_array(self) -> np.ndarray:
        """Array of weights for each proposal"""
        return np.fromiter(self._weights.values(), dtype=float)

    @property
    def n_proposals(self) -> int:
        """Current number of proposals in the meta proposal"""
        return len(self.weights)

    @property
    def flow_config(self) -> dict:
        """Return the configuration for the flow"""
        return self._flow_config

    @flow_config.setter
    def flow_config(self, config: dict) -> None:
        """Set configuration (includes checking defaults)"""
        if config is None:
            config = dict(model_config=dict())
        elif "model_config" not in config:
            config["model_config"] = dict()
        config["model_config"]["n_inputs"] = self.model.dims
        self._flow_config = update_config(config)

    @property
    def _reset_flow(self) -> bool:
        """Boolean to indicate if the flow should be reset"""
        if not self.reset_flow or self.level_count % self.reset_flow:
            return False
        else:
            return True

    @staticmethod
    def _check_fields():
        """Check that the logQ and logW fields have been added."""
        if "logQ" not in config.livepoints.non_sampling_parameters:
            raise RuntimeError(
                "logQ field missing in non-sampling parameters."
            )
        if "logW" not in config.livepoints.non_sampling_parameters:
            raise RuntimeError(
                "logW field missing in non-sampling parameters."
            )

    def initialise(self):
        """Initialise the proposal"""
        self._check_fields()
        if self.initialised:
            logger.debug("Proposal already initialised")
            return

        self.verify_rescaling()

        self.flow = ImportanceFlowModel(
            config=self.flow_config, output=self.output
        )
        self.flow.initialise()
        super().initialise()

    def verify_rescaling(
        self, n: int = 1000, rtol: float = 1e-08, atol: float = 1e-08
    ) -> None:
        """Verify the rescaling is invertible.

        Uses :code:`numpy.allclose`, see numpy documentation for more details.

        Parameters
        ----------
        n : int
            Number of samples to test.
        atol : float
            The absolute tolerance.
        rtol : float
            The relative tolerance.
        """
        logger.debug("Verifying rescaling")
        x_in = self.model.sample_unit_hypercube(n)

        x_prime, log_j = self.rescale(x_in)
        x_re, log_j_inv = self.inverse_rescale(x_prime)

        try:
            assert_structured_arrays_equal(x_re, x_in, atol=atol, rtol=rtol)
        except AssertionError as e:
            raise RuntimeError(f"Rescaling is not invertible. Error: {e}")

        if not np.allclose(log_j, -log_j_inv, atol=atol, rtol=rtol):
            raise RuntimeError(
                "Forward and inverse Jacobian determinants are not equal"
            )
        logger.debug("Rescaling functions are invertible")

    def to_prime(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Convert samples from the unit hypercube to samples in x'-space

        Parameters
        ----------
        x_prime :
            Unstructured array of samples in the unit hypercube

        Returns
        -------
        x :
            Unstructured array of samples in x'-space
        log_j :
            Corresponding log-Jacobian determinant.
        """
        x = np.atleast_2d(x)
        if self.reparameterisation == "logit":
            x_prime, log_j = logit(x, eps=config.general.eps)
            log_j = log_j.sum(axis=1)
        elif self.reparameterisation is None:
            x_prime = x.copy()
            log_j = np.zeros(x.shape[0])
        else:
            raise ValueError(
                f"Unknown reparameterisation: '{self.reparameterisation}'"
            )
        return x_prime, log_j

    def from_prime(self, x_prime: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Convert samples the x'-space to samples in the unit hypercube.

        Parameters
        ----------
        x_prime :
            Unstructured array of samples.

        Returns
        -------
        x :
            Unstructured array of samples in the unit hypercube.
        log_j :
            Corresponding log-Jacobian determinant.
        """
        x_prime = np.atleast_2d(x_prime)
        if self.reparameterisation == "logit":
            x, log_j = sigmoid(x_prime)
            log_j = log_j.sum(axis=1)
        elif self.reparameterisation is None:
            x = x_prime.copy()
            log_j = np.zeros(x.shape[0])
        else:
            raise ValueError(
                f"Unknown reparameterisation: '{self.reparameterisation}'"
            )
        return x, log_j

    def rescale(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Convert to the space in which the flow is trained.

        Returns an unstructured array.
        """
        x = live_points_to_array(x, self.model.names)
        x_prime, log_j = self.to_prime(x)
        return x_prime, log_j

    def inverse_rescale(self, x_prime: np.ndarray) -> np.ndarray:
        """Convert from the space in which the flow is trained.

        Returns a structured array.
        """
        x, log_j = self.from_prime(x_prime)
        if self.clip:
            x = np.clip(x, 0.0, 1.0)
        x = numpy_array_to_live_points(x, self.model.names)
        return x, log_j

    def update_proposal_weights(self, weights: dict) -> None:
        """Method to update the proposal weights dictionary.

        Raises
        ------
        RuntimeError
            If the weights do not sum to 1 are the update.
        """
        self._weights.update(weights)
        w_sum = np.sum(np.fromiter(self._weights.values(), float))
        if not np.isclose(w_sum, 1.0):
            raise RuntimeError(f"Weights must sum to 1! Actual value: {w_sum}")

    def train(
        self,
        samples: np.ndarray,
        plot: bool = False,
        output: Union[str, None] = None,
        weights: np.ndarray = None,
        **kwargs,
    ) -> None:
        """Train the proposal with a set of samples.

        Parameters
        ----------
        samples :  numpy.ndarray
            Array of samples for training.
        plot : bool
            Flag to enable or disable plotting.
        output : Union[str, None]
            Output directory to use instead of default output. If None the
            default that was set when the class what initialised is used.
        kwargs :
            Key-word arguments passed to \
                :py:meth:`nessai.flowmodel.FlowModel.train`.
        """
        self.level_count += 1
        self._weights[self.level_count] = np.nan
        output = self.output if output is None else output
        level_output = os.path.join(output, f"level_{self.level_count}", "")

        if not os.path.exists(level_output):
            os.makedirs(level_output, exist_ok=True)

        training_data = samples.copy()
        x_prime, _ = self.rescale(training_data)

        if plot:
            plot_live_points(
                training_data,
                filename=os.path.join(level_output, "training_data.png"),
            )
            plot_1d_comparison(
                x_prime,
                convert_to_live_points=True,
                filename=os.path.join(level_output, "prime_training_data.png"),
            )

        logger.debug(
            f"Training data min and max: {x_prime.min()}, {x_prime.max()}"
        )

        if self.weighted_kl or weights is not None:
            logger.debug("Using weights in training")
            if weights is not None:
                weights = weights / np.sum(weights)
            else:
                log_weights = training_data["logW"].copy()
                log_weights -= logsumexp(log_weights)
                weights = np.exp(log_weights)
            if np.isnan(weights).any():
                raise ValueError("Weights contain NaN(s)")
            if not np.isfinite(weights).all():
                raise ValueError("Weights contain Inf(s)")

            if plot:
                plot_histogram(
                    weights, filename=level_output + "training_weights.png"
                )
        else:
            weights = None

        self.flow.add_new_flow(reset=self._reset_flow)

        logger.debug(f"Training with {x_prime.shape[0]} samples")
        self.flow.train(
            x_prime,
            weights=weights,
            output=level_output,
            plot=plot or self.plot_training,
            **kwargs,
        )

        if plot:
            test_samples_prime, log_prob = self.flow.sample_and_log_prob(2000)
            test_samples, log_j_inv = self.inverse_rescale(test_samples_prime)
            log_prob -= log_j_inv
            test_samples["logQ"] = log_prob
            plot_live_points(
                test_samples,
                filename=os.path.join(level_output, "generated_samples.png"),
            )

    def compute_log_Q(
        self,
        x_prime: np.ndarray,
        log_j: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Compute the log meta proposal (log Q) for an array of points.

        Parameters
        ----------
        x_prime : numpy.ndarray
            Array of samples in the unit hypercube.
        log_j : Optional[numpy.ndarray]
            Log-Jacobian determinant of the prime samples. Must be supplied if
            proposal includes flows.

        Returns
        -------
        log_Q : numpy.ndarray
            Value of the meta-proposal in the prime space.
        log_q : numpy.ndarray
            Array of values for flow in the flow in the proposal. Array will
            have shape (# samples, # flows)
        """
        if np.isnan(x_prime).any():
            logger.warning("NaNs in samples when computing log_Q")
        if not np.isfinite(x_prime).all():
            logger.warning(
                "Infinite values in the samples when computing log_Q"
            )

        if any(np.isnan(w) for w in self.weights.values()):
            raise RuntimeError("Some weights are not set!")

        log_q_all = np.zeros([x_prime.shape[0], self.n_proposals])
        n_flows = self.flow.n_models

        if self.n_proposals > 1 and log_j is None:
            raise RuntimeError(
                "Must specify log_j! Meta-proposal includes flows"
            )

        if any([flow.training for flow in self.flow.models]):
            raise RuntimeError("One or more flows are in training mode!")

        if n_flows >= 1:
            log_q_all[:, 1 : (n_flows + 1)] = (
                self.flow.log_prob_all(x_prime) + log_j[:, np.newaxis]
            )
        assert log_q_all.shape[0] == x_prime.shape[0]

        logger.debug(f"log_q is nan: {np.isnan(log_q_all).any()}")
        logger.debug(
            f"Mean log q for each each flow: {log_q_all.mean(axis=0)}"
        )
        log_Q = logsumexp(log_q_all, b=self.weights_array, axis=1)

        if np.isnan(log_Q).any():
            raise ValueError("There is a NaN in log g!")

        return log_Q, log_q_all

    def draw(
        self,
        n: int,
        flow_number: Optional[int] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Draw n new points.

        Parameters
        ----------
        n : int
            Number of points to draw.
        flow_number : Optional[int]
            Specifies which flow to use. If not specified the last flow will
            be used.

        Returns
        -------
        np.ndarray :
            Array of new points.
        np.ndarray :
            Log-proposal probabilities (log_q)
        """
        if flow_number is None:
            flow_number = self.level_count

        # Draw a few more samples in case some are not accepted.
        n_draw = int(1.01 * n)
        logger.debug(f"Drawing {n} points")
        samples = np.zeros(0, dtype=self.dtype)
        log_q_samples = np.empty([0, self.n_proposals])

        n_accepted = 0
        while n_accepted < n and n_draw > 0:
            logger.debug(f"Drawing batch of {n_draw} samples")
            # x_prime, log_q = self.flow.sample_and_log_prob(N=n_draw)
            x_prime = self.flow.sample_ith(i=flow_number, N=n_draw)
            log_q = np.ones(n_draw)
            x, log_j_inv = self.inverse_rescale(x_prime)
            # Rescaling can sometimes produce infs that don't appear in samples
            x_check, log_j = self.rescale(x)
            # Probably don't need all these checks.
            acc = (
                self.model.in_unit_hypercube(x)
                & np.isfinite(x_check).all(axis=1)
                & np.isfinite(x_prime).all(axis=1)
                & np.isfinite(log_j)
                & np.isfinite(log_j_inv)
                & np.isfinite(log_q)
            )
            logger.debug(f"Rejected {n_draw - acc.sum()} points")
            if not np.any(acc):
                continue
            x, x_prime, log_j, log_q = get_subset_arrays(
                acc, x, x_prime, log_j, log_q
            )

            x["logQ"], log_q_all = self.compute_log_Q(x_prime, log_j=log_j)
            x["logP"] = self.model.batch_evaluate_log_prior(
                x, unit_hypercube=True
            )
            x["logW"] = -x["logQ"]
            accept = (
                np.isfinite(x["logP"])
                & ~np.isposinf(x["logW"])
                & ~np.isnan(log_q_all).all(axis=1)
                & ~np.isposinf(log_q_all).all(axis=1)
            )
            if not np.any(accept):
                continue

            x, log_q_all = get_subset_arrays(accept, x, log_q_all)

            samples = np.concatenate([samples, x])
            log_q_samples = np.concatenate([log_q_samples, log_q_all], axis=0)

            n_accepted += x.size
            logger.debug(f"Accepted: {n_accepted}")

        samples = samples[:n]
        log_q_samples = log_q_samples[:n]

        logger.debug(f"Returning {samples.size} samples")
        return samples, log_q_samples

    def update_log_q(
        self,
        samples: np.ndarray,
        log_q: np.ndarray,
    ) -> np.ndarray:
        """Update the array of proposal probabilities for a set of samples"""
        if log_q.shape[1] == self.n_proposals:
            raise ValueError("log_q array already contains current proposal")
        x, log_j = self.rescale(samples)
        log_prob_fn = self.get_proposal_log_prob(self.level_count)
        log_q_current = log_prob_fn(x)
        log_q = np.concatenate(
            [log_q, log_q_current[:, np.newaxis] + log_j[:, np.newaxis]],
            axis=1,
        )
        return log_q

    def compute_meta_proposal_from_log_q(self, log_q):
        """Compute the meta-proposal from an array of proposal \
        log-probabilities
        """
        return logsumexp(
            log_q,
            b=self.weights_array,
            axis=1,
        )

    def compute_meta_proposal_samples(self, samples: np.ndarray) -> np.ndarray:
        """Compute the meta proposal Q for a set of samples.

        Includes any rescaling that has been configured.

        Returns
        -------
        log_meta_proposal : numpy.ndarray
            Array of meta-proposal log probabilities (log Q)
        log_q : numpy.ndarray
            Array of log q for each flow.
        """
        if self.level_count not in self.weights or np.isnan(
            self.weights[self.level_count]
        ):
            raise RuntimeError(
                "Weight(s) missing or not set. "
                f"Current weights: {self.weights}."
            )
        x, log_j = self.rescale(samples)
        return self.compute_log_Q(x, log_j=log_j)

    def _log_prior(self, x: np.ndarray) -> np.ndarray:
        """Helper function that returns the prior in the unit hyper-cube."""
        return np.zeros(x.shape[0])

    def get_proposal_log_prob(self, it: int) -> Callable:
        """Get a pointer to the function for ith proposal."""
        if it == -1:
            return self._log_prior
        elif it < len(self.flow.models):
            return lambda x: self.flow.log_prob_ith(x, it)
        else:
            raise ValueError

    def compute_kl_between_proposals(
        self,
        x: np.ndarray,
        p_it: Optional[int] = None,
        q_it: Optional[int] = None,
    ) -> float:
        """Compute the KL divergence between two proposals.

        Samples should be drawn from p. If proposals aren't specified the
        current and previous proposals are used.
        """
        x_prime, log_j = self.rescale(x)
        if p_it is None:
            p_it = self.flow.n_models - 1

        if q_it is None:
            q_it = self.flow.n_models - 2

        if p_it == q_it:
            raise ValueError("p and q must be different")
        elif p_it < -1 or q_it < -1:
            raise ValueError(f"Invalid p_it or q_it: {p_it}, {q_it}")

        log_p_f = self.get_proposal_log_prob(p_it)
        log_q_f = self.get_proposal_log_prob(q_it)

        log_p = log_p_f(x_prime)
        log_q = log_q_f(x_prime)

        if p_it > -1:
            log_p += log_j
        if q_it > -1:
            log_q += log_j

        kl = np.mean(log_p - log_q)
        logger.info(f"KL between {p_it} and {q_it} is: {kl:.3}")
        return kl

    def draw_from_prior(self, n: int) -> Tuple[np.ndarray, np.ndarray]:
        """Draw from the prior"""
        samples = self.model.sample_unit_hypercube(n)
        prime_samples, log_j = self.rescale(samples)
        log_Q, log_q = self.compute_log_Q(prime_samples, log_j=log_j)
        samples["logQ"] = log_Q
        samples["logW"] = -log_Q
        return samples, log_q

    def draw_from_flows(
        self,
        n: int,
        weights=None,
        counts=None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Draw n points from all flows (g).

        Parameters
        ----------
        n : int
            Number of points
        weights
            Array of (normalised) weights. Ignored if counts is specified.
        counts
            Number of samples to draw from each proposal.
        """
        logger.debug(
            f"Drawing {n} samples from the combination of all the proposals"
        )
        if counts is None:
            if weights is None:
                weights = self.weights
            if not np.sum(weights) == 1:
                weights = weights / np.sum(weights)
            if not len(weights) == self.n_proposals:
                ValueError(
                    "Size of weights does not match the number of levels"
                )
            logger.debug(f"Proposal weights: {weights}")
            counts = np.random.multinomial(n, weights)
        else:
            counts = np.array(counts, dtype=int)
            weights = counts / counts.sum()
        logger.debug(f"Expected counts: {counts}")
        if np.any(counts) < 0:
            raise ValueError("Cannot have negative counts")
        if np.sum(counts) == 0:
            raise ValueError("Total counts is zero")
        proposal_id = np.arange(weights.size) - 1
        prime_samples = np.empty([n, self.model.dims])
        sample_its = np.empty(n, dtype=config.livepoints.it_dtype)
        count = 0
        # Draw from prior
        for id, m in zip(proposal_id, counts):
            if m == 0:
                continue
            logger.debug(f"Drawing {m} samples from the {id}th proposal.")
            if id == -1:
                prime_samples[count : (count + m)] = self.to_prime(
                    np.random.rand(m, self.model.dims)
                )[0]
            else:
                prime_samples[count : (count + m)] = self.flow.sample_ith(
                    id, N=m
                )
            sample_its[count : (count + m)] = id
            count += m

        samples, log_j_inv = self.inverse_rescale(prime_samples)
        samples["it"] = sample_its
        x_check, log_j = self.rescale(samples)
        # Probably don't need all these checks.
        finite = (
            self.model.in_unit_hypercube(samples)
            & np.isfinite(x_check).all(axis=1)
            & np.isfinite(prime_samples).all(axis=1)
            & np.isfinite(log_j)
            & np.isfinite(log_j_inv)
        )

        samples, prime_samples, log_j = get_subset_arrays(
            finite, samples, prime_samples, log_j
        )

        log_q = np.zeros((samples.size, self.n_proposals))
        logger.debug("Computing log_q")
        if self.n_proposals > 1:
            log_q[:, 1:] = (
                self.flow.log_prob_all(prime_samples) + log_j[:, np.newaxis]
            )

        # -inf is okay since this is just zero, so only remove +inf or NaN
        finite = ~np.isnan(log_q).all(axis=1) & ~np.isposinf(log_q).all(axis=1)
        samples, log_q = get_subset_arrays(finite, samples, log_q)

        logger.debug(
            f"Mean g for each each flow: {np.exp(log_q).mean(axis=0)}"
        )
        logger.debug(f"Mean log_q for each each flow: {log_q.mean(axis=0)}")

        samples["logP"] = self.model.batch_evaluate_log_prior(
            samples, unit_hypercube=True
        )
        samples, log_q = get_subset_arrays(
            np.isfinite(samples["logP"]), samples, log_q
        )
        counts = np.bincount(
            samples["it"] + 1,
            minlength=self.n_proposals,
        ).astype(int)
        logger.debug(f"Actual counts: {counts}")

        return samples, log_q, counts

    def resume(self, model, flow_config, weights_path=None):
        """Resume the proposal"""
        super().resume(model)
        self.flow_config = flow_config
        self.initialise()
        self.flow.resume(
            self.flow_config["model_config"], weights_path=weights_path
        )

    def __getstate__(self):
        d = self.__dict__
        exclude = {"model", "_flow_config", "flow"}
        state = {k: d[k] for k in d.keys() - exclude}
        return state, self.flow

    def __setstate__(self, state):
        self.__dict__.update(state[0])
        self.flow = state[1]



# -*- coding: utf-8 -*-
"""
Tests for utilities related to python structures such as lists.
"""
import numpy as np
import pytest

from nessai.utils.structures import (
    array_split_chunksize,
    get_inverse_indices,
    get_subset_arrays,
    isfinite_struct,
    replace_in_list,
)


def test_replace_in_list():
    """
    Test if the list produced contains the correct entries in the correct
    locations
    """
    x = [1, 2, 3]
    replace_in_list(x, [1, 2], [5, 4])
    assert x == [5, 4, 3]


def test_replace_in_list_item():
    """
    Test if items are correctly converted to lists in replace_in_list function
    """
    x = [1, 2, 3]
    replace_in_list(x, 3, 4)
    assert x == [1, 2, 4]


def test_replace_in_list_single():
    """Test the function with a list of strings"""
    x = ["aa", "bb", "cc"]
    replace_in_list(x, ["aa", "bb"], ["dd", "ee"])
    assert x == ["dd", "ee", "cc"]


def test_replace_in_list_single_str():
    """Test the function with single strings"""
    x = ["aa", "bb", "cc"]
    replace_in_list(x, "aa", "dd")
    assert x == ["dd", "bb", "cc"]


def test_different_lengths():
    """
    Assert an error is raised if the targets and replacements are different
    lengths.
    """
    with pytest.raises(RuntimeError) as excinfo:
        replace_in_list([1, 2], [1, 2], 3)
    assert "Targets and replacements are different lengths!" in str(
        excinfo.value
    )


def test_missing_targets():
    """Assert an error is raised if a target is not in the target list."""
    with pytest.raises(ValueError) as excinfo:
        replace_in_list([1, 2], 4, 3)
    assert "Targets [4] not in list: [1, 2]" in str(excinfo.value)


def test_get_subset_arrays():
    """Assert the correct subsets are returned."""
    a = np.array([1, 2, 3])
    b = np.array([4, 5, 6])
    indices = np.array([1, 2])
    a_out, b_out = get_subset_arrays(indices, a, b)
    np.testing.assert_equal(a_out, a[indices])
    np.testing.assert_equal(b_out, b[indices])


def test_get_subset_arrays_empty():
    """Assert output is empty if no arrays are provided"""
    out = get_subset_arrays(np.array([1, 2]))
    assert out == ()


@pytest.mark.parametrize(
    "names, expected",
    [
        (None, [False, True, False]),
        (["x", "y", "z"], [False, True, False]),
        (["x"], [True, True, True]),
        (["y"], [False, True, True]),
        (["z"], [True, True, False]),
    ],
)
def test_isfinite_struct(names, expected):
    """Assert the correct array is returned"""
    x = np.array(
        [(0, np.inf, 0), (1, 1, 1), (2, 2, np.nan)],
        dtype=[("x", "f8"), ("y", "f8"), ("z", "f8")],
    )
    out = isfinite_struct(x, names=names)
    assert len(out) == 3
    np.testing.assert_equal(out, np.array(expected))


@pytest.mark.integration_test
def test_isfinite_struct_invalid_name():
    """Assert an error is raised if a name is invalid.

    Numpy should raise a ValueError if the name is not a field in the array.
    """
    x = np.array([(1,), (2,)], dtype=[("x", "i4")])
    with pytest.raises(ValueError):
        isfinite_struct(x, ["y"])


def test_array_split_chunksize():
    """Assert the correct array sizes are returned"""
    a = np.array([1, 2, 3, 4, 5])
    out = array_split_chunksize(a, 2)
    assert len(out) == 3
    np.testing.assert_array_equal(out[0], a[:2])
    np.testing.assert_array_equal(out[1], a[2:4])
    np.testing.assert_array_equal(out[2], a[4:])


def test_array_split_chunksize_larger_than_array():
    """Assert the correct array sizes are returned"""
    a = np.array([1, 2, 3, 4, 5])
    out = array_split_chunksize(a, 6)
    assert len(out) == 1
    np.testing.assert_array_equal(out[0], a)


def test_array_split_chunksize_invalid_chunksize():
    """Assert an error is returned if the chunksize is less than one"""
    with pytest.raises(ValueError, match="chunksize must be greater than 1"):
        array_split_chunksize(np.array([1, 2]), -1)


def test_get_inverse_indices():
    """Assert the correct indices are returned"""
    indices = np.array([1, 2, 3])
    n = 5
    expected = np.array([0, 4])
    out = get_inverse_indices(n, indices)
    np.testing.assert_array_equal(out, expected)


def test_get_inverse_indices_empty_inverse():
    """Assert an empty array is returned in the indices are complete"""
    indices = np.array([0, 1, 2, 3, 4])
    n = 5
    expected = np.array([])
    out = get_inverse_indices(n, indices)
    np.testing.assert_array_equal(out, expected)


def test_get_inverse_indices_invalid_n():
    """Assert"""
    indices = np.array([0, 1, 4])
    n = 4
    with pytest.raises(
        ValueError, match=r"Indices contain values that are out of range for n"
    ):
        get_inverse_indices(n, indices)



# -*- coding: utf-8 -*-
"""End-to-end test of the populate method"""
import numpy as np
import pytest
import torch

from nessai.proposal import FlowProposal
from nessai.livepoint import numpy_array_to_live_points
from nessai.utils.testing import assert_structured_arrays_equal

torch.set_num_threads(1)


@pytest.mark.parametrize()
def flow_config():
    return dict(
        model_config=dict(n_neurons=1, n_blocks=1, n_layers=1, kwargs={})
    )


@pytest.mark.parametrize("expansion_fraction", [0.0, 1.0, None])
@pytest.mark.parametrize("check_acceptance", [False, True])
@pytest.mark.parametrize("rescale_parameters", [False, True])
@pytest.mark.integration_test
@pytest.mark.timeout(30)
def test_flowproposal_populate(
    tmp_path,
    model,
    flow_config,
    expansion_fraction,
    check_acceptance,
    rescale_parameters,
):
    """
    Test the populate method in the FlowProposal class with a range of
    parameters
    """
    output = tmp_path / "flowproposal"
    output.mkdir()
    n_draw = 10
    fp = FlowProposal(
        model,
        output=output,
        flow_config=flow_config,
        plot=False,
        poolsize=10,
        latent_prior="truncated_gaussian",
        expansion_fraction=expansion_fraction,
        check_acceptance=check_acceptance,
        rescale_parameters=rescale_parameters,
        max_radius=1.0,
        constant_volume_mode=False,
    )

    fp.initialise()
    worst = numpy_array_to_live_points(0.01 * np.ones(fp.dims), fp.names)
    fp.populate(worst, N=n_draw)

    assert fp.x.size == n_draw


@pytest.mark.parametrize(
    "latent_prior",
    [
        "gaussian",
        "truncated_gaussian",
        "uniform_nball",
        "uniform_nsphere",
        "uniform",
    ],
)
@pytest.mark.integration_test
@pytest.mark.timeout(30)
def test_flowproposal_populate_edge_cases(
    tmp_path, model, flow_config, latent_prior
):
    """Tests some less common settings for flowproposal"""
    output = tmp_path / "flowproposal"
    output.mkdir()
    n_draw = 2
    fp = FlowProposal(
        model,
        output=output,
        flow_config=flow_config,
        plot=False,
        poolsize=10,
        latent_prior=latent_prior,
        expansion_fraction=None,
        rescale_parameters=False,
        max_radius=0.1,
        constant_volume_mode=False,
    )

    fp.initialise()
    assert fp.rescale_parameters is False
    assert fp.parameters_to_rescale == []
    worst = numpy_array_to_live_points(0.01 * np.ones(fp.dims), fp.names)
    fp.populate(worst, N=n_draw)

    assert fp.x.size == n_draw


@pytest.mark.parametrize("plot", [False, True])
@pytest.mark.integration_test
def test_training(tmpdir, model, plot):
    """Integration test to test training the flow with and without plotting."""
    output = str(tmpdir.mkdir("test_train"))
    config = dict(max_epochs=10)
    fp = FlowProposal(
        model, output=output, plot="min", poolsize=100, flow_config=config
    )

    fp.initialise()

    x = model.new_point(500)
    fp.train(x, plot=plot)

    assert fp.training_count == 1
    assert fp.populated is False


@pytest.mark.parametrize("check_acceptance", [False, True])
@pytest.mark.parametrize("rescale_parameters", [False, True])
@pytest.mark.integration_test
@pytest.mark.timeout(30)
def test_constant_volume_mode(
    tmpdir, model, flow_config, check_acceptance, rescale_parameters
):
    """Integration test for constant volume mode.

    With q=0.8647 should get a radius of ~2.
    """
    output = str(tmpdir.mkdir("flowproposal"))
    expected_radius = 2.0
    fp = FlowProposal(
        model,
        output=output,
        flow_config=flow_config,
        plot=False,
        poolsize=10,
        constant_volume_mode=True,
        volume_fraction=0.8647,
        check_acceptance=check_acceptance,
        rescale_parameters=rescale_parameters,
    )
    fp.initialise()
    worst = numpy_array_to_live_points(0.5 * np.ones(fp.dims), fp.names)
    fp.populate(worst, N=10)
    assert fp.x.size == 10

    np.testing.assert_approx_equal(fp.r, expected_radius, 4)
    np.testing.assert_approx_equal(fp.fixed_radius, expected_radius, 4)


@pytest.mark.parametrize("boundary_inversion", [True, False])
@pytest.mark.parametrize("inversion_type", ["split", "duplicate"])
@pytest.mark.integration_test
def test_verify_rescaling_integration(
    tmp_path, model, boundary_inversion, inversion_type
):
    """Assert verify rescaling passes."""
    output = tmp_path / "test"
    output.mkdir()

    fp = FlowProposal(
        model,
        output=output,
        poolsize=10,
        boundary_inversion=boundary_inversion,
        inversion_type=inversion_type,
    )
    fp.set_rescaling()
    fp.verify_rescaling()


@pytest.mark.integration_test
def test_rescaling_integration_with_rescale_parameters(tmp_path, model):
    """Assert rescale parameters is configured correctly"""
    output = tmp_path / "test"
    output.mkdir()

    fp = FlowProposal(
        model,
        output=output,
        poolsize=10,
        rescale_parameters=True,
        rescale_bounds=[0, 1],
    )
    fp.set_rescaling()

    n = 10
    x = model.new_point(n)

    x_prime, log_j = fp.rescale(x)

    x_recon, log_j_inv = fp.inverse_rescale(x_prime)

    x_prime_expected = x_prime.copy()
    for name in model.names:
        x_prime_expected[name + "_prime"] = (
            x[name] - model.bounds[name][0]
        ) / (model.bounds[name][1] - model.bounds[name][0])
    expected_log_j = -np.log(np.prod(model.upper_bounds - model.lower_bounds))

    assert len(log_j) == n
    assert (log_j == expected_log_j).all()
    np.testing.assert_array_equal(log_j, -log_j_inv)

    assert_structured_arrays_equal(x_prime, x_prime_expected)
    assert_structured_arrays_equal(x_recon, x)


@pytest.mark.integration_test
def test_rescaling_integration_no_rescaling(tmp_path, model):
    """Assert setting rescale_parameters=False does not rescale any
    parameters.
    """
    output = tmp_path / "test"
    output.mkdir()

    fp = FlowProposal(
        model,
        output=output,
        poolsize=10,
        rescale_parameters=False,
    )
    fp.set_rescaling()

    n = 10
    x = model.new_point(n)

    x_prime, log_j = fp.rescale(x)

    x_recon, log_j_inv = fp.inverse_rescale(x_prime)

    assert len(log_j) == n
    assert (log_j == 0.0).all()
    np.testing.assert_array_equal(log_j, -log_j_inv)

    assert_structured_arrays_equal(x_prime, x)
    assert_structured_arrays_equal(x_recon, x_prime)



# -*- coding: utf-8 -*-
"""
Tests for the stats related utilities.
"""
import numpy as np
import pytest

from nessai.utils.stats import (
    effective_sample_size,
    rolling_mean,
    weighted_quantile,
)


def test_ess():
    """Test the effective sample size"""
    log_w = np.zeros(10)
    np.testing.assert_almost_equal(effective_sample_size(log_w), 10)
    # Make the input array remains unchanged
    assert (log_w == 0.0).all()


def test_rolling_mean():
    """Test the rolling mean."""
    x = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
    expected = np.array([4.0 / 3.0, 2.0, 3.0, 4.0, 5.0, 17.0 / 3.0])
    out = rolling_mean(x, N=3)
    np.testing.assert_array_almost_equal(out, expected, decimal=15)


def test_weighted_quantile_equal_weights():
    """Test the weighted quanitle method"""
    x = [1, 2, 3, 4, 5]
    quantile = 0.5
    out = weighted_quantile(x, quantile)
    assert out == 3.0


def test_weighted_quantile_different_weights():
    """Test the weighted quantile method.

    Based on example 8 from https://aakinshin.net/posts/weighted-quantiles/
    """
    x = [1, 2, 3, 4, 5]
    w = [0.4, 0.4, 0.05, 0.05, 0.1]
    quantile = 0.5
    out = weighted_quantile(x, quantile, log_weights=np.log(w))
    np.testing.assert_almost_equal(out, 1.8416, decimal=4)


def test_weighted_quantile_value_error_quantile():
    """Assert an error is raised if the quantiles are invalid"""
    with pytest.raises(ValueError, match=r"Quantiles should be in \[0, 1\]"):
        weighted_quantile([1, 2, 3], quantiles=1.5)


def test_weighted_quantile_value_error_neff():
    """Assert an error is raised if the ESS is not finite."""
    with pytest.raises(
        ValueError, match=r"Effective sample size is not finite.*"
    ):
        weighted_quantile(
            [1, 2], 0.5, log_weights=np.array([np.NINF, np.NINF])
        )



import matplotlib.pyplot as plt
from nessai.samplers.importancesampler import ImportanceNestedSampler as INS
import numpy as np
import pytest


@pytest.fixture(autouse=True)
def auto_close_figures():
    """Automatically close all figures after each test"""
    yield
    plt.close("all")


def test_plot_state(ins, history, n_it):
    ins.iteration = n_it
    ins.history = history
    ins.history["checkpoint_iterations"] = [3, 4]
    ins.importance = dict(
        total=np.arange(-1, n_it),
        evidence=np.arange(-1, n_it),
        posterior=np.arange(-1, n_it),
    )
    ins.stopping_criterion = ["ratio", "ess"]
    ins.tolerance = [0.0, 1000]
    fig = INS.plot_state(ins)
    assert fig is not None


def test_plot_extra_state(ins, history, n_it):
    ins.iteration = n_it
    ins.checkpoint_iterations = [3, 4]
    ins.history = history
    fig = INS.plot_extra_state(ins)
    assert fig is not None


@pytest.mark.parametrize("enable_colours", [False, True])
def test_plot_trace(ins, samples, enable_colours):
    ins.samples_unit = samples
    fig = INS.plot_trace(ins, enable_colours=enable_colours)
    assert fig is not None


def test_plot_likelihood_levels(ins, samples):
    ins.samples_unit = samples
    fig = INS.plot_likelihood_levels(ins)
    assert fig is not None


def test_plot_level_cdf(ins, samples):
    cdf = np.cumsum(samples["logW"])
    fig = INS.plot_level_cdf(
        ins,
        samples["logL"],
        cdf,
        q=0.5,
        threshold=samples[len(samples) // 2]["logL"],
    )
    assert fig is not None



# -*- coding: utf-8 -*-
"""
Test the analytic proposal method.
"""
import datetime
import numpy as np
import pytest
from unittest.mock import MagicMock, Mock, create_autospec, patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.proposal import AnalyticProposal


@pytest.fixture
def proposal():
    return create_autospec(AnalyticProposal)


def test_init(proposal):
    """Test the init method."""
    with patch("nessai.proposal.analytic.Proposal.__init__") as mock_super:
        AnalyticProposal.__init__(proposal, "model", poolsize=10, test=True)
    mock_super.assert_called_once_with("model", test=True)
    assert proposal._poolsize == 10
    assert proposal.populated is False


def test_poolsize(proposal):
    """Test poolsize property"""
    proposal._poolsize = 100
    poolsize = AnalyticProposal.poolsize.__get__(proposal)
    assert poolsize == 100


@pytest.mark.parametrize("N", [None, 5])
def test_populate(proposal, N):
    """Test the populate process"""
    poolsize = 10
    if N is None:
        samples = numpy_array_to_live_points(
            np.arange(poolsize)[:, np.newaxis], "x"
        )
        log_p = np.arange(poolsize, 2 * poolsize)
    else:
        samples = numpy_array_to_live_points(np.arange(N)[:, np.newaxis], "x")
        log_p = np.arange(N, 2 * N)
    log_l = np.random.rand(samples.size)
    proposal.poolsize = poolsize
    proposal.model = Mock()
    proposal.model.new_point = Mock(return_value=samples)
    proposal.model.batch_evaluate_log_prior = Mock(return_value=log_p)
    proposal.model.batch_evaluate_log_likelihood = MagicMock(
        return_value=log_l
    )
    AnalyticProposal.populate(proposal, N=N)

    if N is None:
        N = poolsize
    proposal.model.new_point.assert_called_once_with(N=N)
    proposal.model.batch_evaluate_log_prior.assert_called_once_with(samples)
    proposal.model.batch_evaluate_log_likelihood.assert_called_once_with(
        proposal.samples
    )

    np.testing.assert_array_equal(proposal.samples["logP"], log_p)
    assert sorted(proposal.indices) == list(range(N))
    assert proposal.populated is True
    np.testing.assert_array_equal(proposal.samples["logL"], log_l)


@pytest.mark.parametrize("populated", [True, False])
def test_draw(proposal, populated, wait):
    """Test the draw method"""
    N = 5
    proposal.populated = populated
    proposal.populate = Mock(side_effect=wait)
    proposal.population_time = datetime.timedelta()
    proposal.indices = [0, 1, 2, 3, 4]
    proposal.samples = np.array([0, 1, 2, 3, 4])

    sample = AnalyticProposal.draw(proposal, 1, N=N)

    assert sample == 4

    if not populated:
        proposal.populate.assert_called_once_with(N=N)
        assert proposal.population_time.total_seconds() > 0.0
    else:
        proposal.populate.assert_not_called()
        assert proposal.population_time.total_seconds() == 0.0
    assert proposal.indices == [0, 1, 2, 3]


def test_draw_out_of_samples(proposal):
    """Assert populated is set to false when the last sample is used."""
    N = 5
    proposal.populated = True
    proposal.indices = [0]
    proposal.samples = np.array([0, 1, 2, 3, 4])

    sample = AnalyticProposal.draw(proposal, 1, N=N)

    assert sample == 0
    assert proposal.indices == []
    assert proposal.populated is False


@pytest.mark.integration_test
def test_draw_intergration(model):
    """Integration test for the draw method"""
    proposal = AnalyticProposal(model)
    N = 10
    old_point = model.new_point()
    assert not proposal.populated
    proposal.draw(old_point, N=N)
    assert proposal.populated
    [proposal.draw(old_point) for _ in range(N - 1)]
    assert not proposal.indices
    assert not proposal.populated


@pytest.mark.integration_test
def test_populate_integration(model):
    """Integration test for the populate method"""
    proposal = AnalyticProposal(model)
    N = 500
    proposal.populate(N=N)
    assert proposal.samples.size == N
    assert proposal.populated is True



# -*- coding: utf-8 -*-
"""
Default configuration of FlowModel.
"""
DEFAULT_MODEL_CONFIG = dict(
    n_inputs=None,
    n_neurons=None,
    n_blocks=4,
    n_layers=2,
    ftype="RealNVP",
    device_tag="cpu",
    flow=None,
    inference_device_tag=None,
    distribution=None,
    distribution_kwargs=None,
    kwargs=dict(batch_norm_between_layers=True, linear_transform="lu"),
)

DEFAULT_FLOW_CONFIG = dict(
    lr=0.001,
    annealing=False,
    clip_grad_norm=5,
    batch_size=1000,
    val_size=0.1,
    max_epochs=500,
    patience=20,
    noise_type=None,
    noise_scale=None,
    use_dataloader=False,
    optimiser="adamw",
    optimiser_kwargs={},
)



from unittest.mock import MagicMock, create_autospec
from nessai.flows.distributions import ResampledGaussian


def test_finalise():
    """Test the finalise method"""

    dist = create_autospec(ResampledGaussian)
    dist.estimate_normalisation_constant = MagicMock()

    ResampledGaussian.finalise(dist, n_samples=100, n_batches=10)

    dist.estimate_normalisation_constant.assert_called_once_with(
        n_samples=100,
        n_batches=10,
    )



"""Test configuration of INS"""

from unittest.mock import MagicMock

from nessai import config as nessai_config
from nessai.samplers.importancesampler import ImportanceNestedSampler as INS
import numpy as np
import pytest


@pytest.mark.parametrize("save_log_q", [False, True])
def test_init(ins, model, save_log_q):
    ins.add_fields = MagicMock()
    INS.__init__(ins, model, save_log_q=save_log_q, draw_iid_live=True)
    ins.add_fields.assert_called_once()
    assert ins.training_samples.save_log_q is save_log_q
    assert ins.iid_samples.save_log_q is save_log_q


def test_add_fields():
    INS.add_fields()
    assert "logW" in nessai_config.livepoints.non_sampling_parameters
    assert "logQ" in nessai_config.livepoints.non_sampling_parameters


@pytest.mark.parametrize("it, expected", [(None, -1), (100, 100)])
def test_configure_min_iterations(ins, it, expected):
    INS.configure_iterations(ins, min_iteration=it)
    assert ins.min_iteration == expected


@pytest.mark.parametrize("it, expected", [(None, np.inf), (100, 100)])
def test_configure_max_iterations(ins, it, expected):
    INS.configure_iterations(ins, max_iteration=it)
    assert ins.max_iteration == expected


def test_initialise(ins):
    ins.initialised = False
    ins.live_points_unit = None
    ins.populate_live_points = MagicMock()
    ins.proposal = MagicMock()

    INS.initialise(ins)

    ins.populate_live_points.assert_called_once()
    ins.proposal.initialise.assert_called_once()
    assert ins.initialised is True


def test_initialise_history(ins):
    ins.history = None
    INS.initialise_history(ins)
    assert ins.history is not None


def test_check_configuration_min_samples(ins):
    ins.min_samples = 100
    ins.nlive = 10
    ins.min_remove = 1
    with pytest.raises(
        ValueError, match="`min_samples` must be less than `nlive`"
    ):
        INS.check_configuration(ins)


def test_check_configuration_min_remove(ins):
    ins.min_samples = 50
    ins.nlive = 100
    ins.min_remove = 200
    with pytest.raises(
        ValueError, match="`min_remove` must be less than `nlive`"
    ):
        INS.check_configuration(ins)


def check_configuration_okay(ins):
    ins.min_samples = 50
    ins.nlive = 100
    ins.min_remove = 1
    assert INS.check_configuration(ins) is True



"""Integration test for the reparameterisation in GWFlowProposal"""

from nessai.gw.proposal import GWFlowProposal
import numpy as np
import pytest


@pytest.mark.requires("bilby")
@pytest.mark.requires("lal")
@pytest.mark.requires("astropy")
@pytest.mark.integration_test
@pytest.mark.parametrize("reverse", [False, True])
def test_invertibility(
    tmp_path, get_bilby_gw_model, injection_parameters, reverse
):
    """Make sure the default GW reparameterisation is invertible.

    Uses the `verify_rescaling` method from `FlowProposal`.

    Also checks if the number of unique samples is the same as the input. This
    will depend on the order of the reparameterisations when using inversion.
    """
    model = get_bilby_gw_model(
        list(injection_parameters.keys()), injection_parameters
    )
    output = tmp_path / "test"
    output.mkdir()

    proposal = GWFlowProposal(
        model=model,
        poolsize=1000,
        output=output,
        reverse_reparameterisations=reverse,
    )

    proposal.set_rescaling()
    proposal.verify_rescaling()

    reparam = proposal._reparameterisation
    assert reparam.order == list(reparam.keys())

    if not reverse:
        pytest.xfail(
            "Checking unique samples with reverse=False will always fail."
        )
    # Check the number of unique samples
    n = 10
    x = model.new_point(n)

    x_prime, _ = proposal.rescale(x)
    x_out, _ = proposal.inverse_rescale(x_prime)

    flags = {n: False for n in proposal.names}
    for name in proposal.names:
        flags[name] = np.unique(x_out[name].round(decimals=10)).size == n
    if not all(flags.values()):
        msg = f"""Number of unique samples has changed.
        Breakdown per parameter: {flags}
        """
        raise AssertionError(msg)



# -*- coding: utf-8 -*-
"""
Utilities related to statistics.
"""
import numpy as np
from scipy.special import logsumexp, betainc


def effective_sample_size(log_w):
    """Compute Kish's effective sample size.

    Parameters
    ----------
    log_w : array_like
        Log-weights.

    Returns
    -------
    float
        The effective sample size.
    """
    log_w = np.array(log_w)
    log_w -= logsumexp(log_w)
    n = np.exp(-logsumexp(2 * log_w))
    return n


def rolling_mean(x, N=10):
    """Compute the rolling mean with a given window size.

    Based on this answer from StackOverflow: \
        https://stackoverflow.com/a/47490020

    Parameters
    ----------
    x : :obj:`numpy.ndarray`
        Array of samples
    N : int
        Size of the window over which the rolling mean is computed.

    Returns
    -------
    :obj:`numpy.ndarray`
        Array containing the moving average.
    """
    # np.cumsum is faster but doesn't work with infs in the data.
    return np.convolve(
        np.pad(x, (N // 2, N - 1 - N // 2), mode="edge"),
        np.ones(N) / N,
        mode="valid",
    )


def weighted_quantile(
    values, quantiles, log_weights=None, values_sorted=False
):
    """Compute quantiles for an array of values.

    Uses the Harrell-Davis quantile estimator.

    Parameters
    ----------
    values : array_like
        Array of values
    quantiles : float or array_like
        Quantiles to compute
    log_weights : array_like, optional
        Array of log-weights
    values_sorted : bool
        If the values are pre-sorted or not

    Returns
    -------
    np.ndarray
        Array of values for each quantile.

    Raises
    ------
    ValueError
        If the effective sample size is not finite.
    """
    values = np.asarray(values)
    quantiles = np.asarray(quantiles)
    if log_weights is None:
        log_weights = np.zeros(len(values))
    log_weights = np.asarray(log_weights)

    if not (np.all(quantiles >= 0) and np.all(quantiles <= 1)):
        raise ValueError("Quantiles should be in [0, 1]")

    if not values_sorted:
        idx = np.argsort(values)
        values = values[idx]
        log_weights = log_weights[idx]

    log_weights = log_weights - logsumexp(log_weights)
    neff = np.exp(-logsumexp(2 * log_weights))
    weights = np.exp(log_weights)
    if not np.isfinite(neff):
        raise ValueError(
            "Effective sample size is not finite, cannot compute weighted "
            "quantile."
        )
    a = quantiles * (neff + 1)
    b = (1 - quantiles) * (neff + 1)

    end_points = np.zeros((len(weights) + 1, 1), dtype=weights.dtype)
    np.cumsum(weights, out=end_points[1:, 0], axis=0)
    end_points /= end_points[-1]

    # Regularised incomplete beta function
    b_li = betainc(a, b, end_points[:-1])
    b_ri = betainc(a, b, end_points[1:])

    w_star = b_ri - b_li
    if values.ndim == 1:
        values = np.expand_dims(values, -1)
    q_p = np.sum(w_star * values, axis=0)

    return q_p



# -*- coding: utf-8 -*-
"""
Plotting utilities.
"""
import functools
import logging
from typing import List, Optional

from cycler import cycler
import matplotlib as mpl
from matplotlib import pyplot as plt
import numpy as np
import seaborn as sns
import pandas as pd

from . import config
from .livepoint import live_points_to_array
from .utils import auto_bins

logger = logging.getLogger(__name__)

_rcparams = sns.plotting_context("notebook")
_rcparams.update(
    {
        "legend.frameon": False,
    }
)


def nessai_style(line_styles=True):
    """Decorator for plotting function that sets the style.

    Functions as both standard decorator :code:`@nessai_style` or as a callable
    decorator :code:`@nessai_style()`.

    Style can be disabled by setting :py:data:`nessai.config.DISABLE_STYLE` to
    :code:`True`.

    Parameters
    ----------
    line_styles : boolean
        Use custom line styles.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if config.plotting.disable_style:
                return func(*args, **kwargs)
            c = cycler(color=config.plotting.line_colours)
            if line_styles:
                c += cycler(linestyle=config.plotting.line_styles)
            d = {
                "axes.prop_cycle": c,
            }
            with sns.axes_style(config.plotting.sns_style), mpl.rc_context(
                {**_rcparams, **d}
            ):
                return func(*args, **kwargs)

        return wrapper

    if callable(line_styles):
        return decorator(line_styles)
    else:
        return decorator


@nessai_style()
def plot_live_points(
    live_points, filename=None, bounds=None, c=None, **kwargs
):
    """
    Plot a set of live points in a corner-like plot.

    Will drop columns where all elements are NaNs

    Parameters
    ----------
    live_points : ndarray
        Structured array of live points to plot.
    filename : str
        Filename for resulting figure
    bounds : dict:
        Dictionary of lower and upper bounds to plot
    c : str, optional
        Name of field in the structured array to use as the hue when plotting
        the samples. If not specified, no hue is used.
    kwargs :
        Keyword arguments used to update the pairplot kwargs. Diagonal and off-
        diagonal plots can be configured with ``diag_kws`` and ``plot_kws``.
    """
    pairplot_kwargs = dict(
        corner=True,
        kind="scatter",
        diag_kws=dict(
            histtype="step",
            bins="auto",
            lw=1.5,
            density=True,
            color=config.plotting.base_colour,
        ),
        plot_kws=dict(
            s=1.0,
            edgecolor=None,
            palette="viridis",
            color=config.plotting.base_colour,
        ),
    )
    pairplot_kwargs.update(kwargs)

    df = pd.DataFrame(live_points)
    df = df.dropna(axis="columns", how="all")
    df = df[np.isfinite(df).all(1)]

    if c is not None:
        hue = df[c]
        if np.all(hue == hue[0]):
            logger.warning(
                f"Selected hue variable: {c} is constant! Disabling."
            )
            hue = None
    else:
        hue = None

    if hue is None:
        pairplot_kwargs["plot_kws"].pop("palette")

    fig = sns.PairGrid(df, corner=True, diag_sharey=False)
    fig.map_diag(plt.hist, **pairplot_kwargs["diag_kws"])
    fig.map_offdiag(sns.scatterplot, hue=hue, **pairplot_kwargs["plot_kws"])

    if bounds is not None:
        for i, v in enumerate(bounds.values()):
            fig.axes[i, i].axvline(
                v[0],
                ls=":",
                alpha=0.5,
                color=config.plotting.highlight_colour,
            )
            fig.axes[i, i].axvline(
                v[1],
                ls=":",
                alpha=0.5,
                color=config.plotting.highlight_colour,
            )

    if filename is not None:
        fig.savefig(filename)
        plt.close()
    else:
        return fig


@nessai_style()
def plot_1d_comparison(
    *live_points,
    parameters=None,
    labels=None,
    colours=None,
    bounds=None,
    hist_kwargs={},
    filename=None,
    convert_to_live_points=False,
):
    """
    Plot 1d histograms comparing different sets of live points

    Parameters
    ----------
    *live_points : iterable of ndarrays
        Variable length argument of live points in structured arrays with
        fields. Also see ``parameters`` argument.
    parameters : array_like, optional
        Array of parameters (field names) to plot. Default None implies all
        fields are plotted.
    labels : list, optional
        List of labels for each structured array being plotted (default None).
        If None each set of live points is labelled numerically
    colours : list, optional
        List of colours to use for each set of live points.
    bounds : dict, optional
        Dictionary of upper and lowers bounds to plot. Each key must
        match a field and each value must be an interable of length 2 in order
        lower then upper bound. If None (default), no bounds plotted.
    hist_kwargs : dict, optional
        Dictionary of keyword arguments passed to matplotlib.pyplot.hist.
    filename : str, optional
        Name of file for saving figure. (Default None implies figure is not
        saved).
    convert_to_live_points : bool, optional
        Set to true if inputs are not structured arrays of live points
    """
    if convert_to_live_points:
        live_points = list(live_points)
        if parameters is None:
            parameters = [i for i in range(live_points[0].shape[-1])]
        for i in range(len(live_points)):
            live_points[i] = {
                k: v for k, v in zip(parameters, live_points[i].T)
            }

    elif any(lp.dtype.names is None for lp in live_points):
        raise RuntimeError(
            "Live points are not structured arrays"
            "Set `convert_to_live_points=True`."
        )
    elif parameters is None:
        parameters = live_points[0].dtype.names

    if labels is None:
        labels = [str(i) for i in range(len(live_points))]
    elif not len(labels) == len(live_points):
        raise ValueError(
            "Length of labels list must match number of arrays being plotted."
        )

    if colours is None:
        colours = sns.color_palette()
        colours = int(np.ceil(len(live_points) / len(colours))) * colours
    elif not len(colours) == len(live_points):
        raise ValueError(
            "Length of colours list must match number of arrays being plotted."
        )

    figsize = (3, min(config.plotting.max_figsize, 3 * len(parameters)))
    fig, axs = plt.subplots(len(parameters), 1, sharey=False, figsize=figsize)

    if len(parameters) > 1:
        axs = axs.ravel()
    else:
        axs = [axs]

    for i, f in enumerate(parameters):

        finite_points = []
        include = []
        for j, lp in enumerate(live_points):
            if np.isnan(lp[f]).all():
                continue
            idx = np.isfinite(lp[f])
            if idx.any():
                finite_points.append(lp[f][idx])
                include.append(j)
        if not include:
            logger.warning(f"No finite points for {f}, skipping.")
            continue

        xmin = np.min([p.min() for p in finite_points])
        xmax = np.max([p.max() for p in finite_points])

        for j, p in enumerate(finite_points):
            orig_idx = include[j]
            axs[i].hist(
                p,
                bins=auto_bins(p),
                histtype="step",
                range=(xmin, xmax),
                density=True,
                label=labels[orig_idx],
                color=colours[orig_idx],
                **hist_kwargs,
            )
        axs[i].set_xlabel(f)
        if bounds is not None and f in bounds:
            axs[i].axvline(bounds[f][0], ls=":", alpha=0.5, color="k")
            axs[i].axvline(bounds[f][1], ls=":", alpha=0.5, color="k")

    if len(labels) > 1:
        # Get the handles
        # Some axes may have less due to e.g. all NaN values, so loop over them
        # all until a valid set of handles is found.
        for ax in axs:
            handles, _ = ax.get_legend_handles_labels()
            if len(handles) == len(labels):
                break
        if len(handles) == len(labels):
            legend_labels = dict(zip(labels, handles))
            fig.legend(
                legend_labels.values(),
                legend_labels.keys(),
                frameon=False,
                ncol=len(labels),
                loc="upper center",
                bbox_to_anchor=(0, 0.1 / len(parameters), 1, 1),
                bbox_transform=plt.gcf().transFigure,
            )
        else:
            logger.warning("Could not plot legend")

    plt.tight_layout()
    if filename is not None:
        fig.savefig(filename, bbox_inches="tight")
        plt.close(fig)
    else:
        return fig


@nessai_style(line_styles=False)
def plot_indices(indices, nlive=None, filename=None, plot_breakdown=True):
    """
    Histogram indices for index insertion tests, also includes the CDF.

    Parameters
    ----------
    indices : array_like
        List of insertion indices to plot
    nlive : int
        Number of live points used in the nested sampling run
    filename : str
        Filename used to save the figure.
    plot_breakdown : bool, optional
       If true, then the CDF for every nlive points is also plotted as grey
       lines.
    """
    indices = np.asarray(indices)
    if not indices.size or not nlive:
        logger.warning("Not producing indices plot.")
        return

    fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    nbins = min(len(np.histogram_bin_edges(indices, "auto")) - 1, 1000)
    if plot_breakdown:
        for i in range(len(indices) // nlive):
            ax[1].hist(
                indices[i * nlive : (i + 1) * nlive],
                bins=nlive,
                histtype="step",
                density=True,
                alpha=0.1,
                color="black",
                lw=0.5,
                cumulative=True,
                range=(0, nlive - 1),
            )

    ax[0].hist(
        indices,
        density=True,
        color="tab:blue",
        linewidth=1.25,
        histtype="step",
        bins=nbins,
        label="produced",
        range=(0, nlive - 1),
    )
    ax[1].hist(
        indices,
        density=True,
        color="tab:blue",
        linewidth=1.25,
        histtype="step",
        bins=nlive,
        label="produced",
        cumulative=True,
        range=(0, nlive - 1),
    )

    if nlive is not None:
        ax[0].axhline(
            1 / nlive,
            color="black",
            linewidth=1.25,
            linestyle="-",
            label="pmf",
            alpha=0.5,
        )
        ax[0].axhline(
            (1 + (nbins / len(indices)) ** 0.5) / nlive,
            color="black",
            linewidth=1.25,
            linestyle=":",
            alpha=0.5,
            label="1-sigma",
        )
        ax[0].axhline(
            (1 - (nbins / len(indices)) ** 0.5) / nlive,
            color="black",
            linewidth=1.25,
            linestyle=":",
            alpha=0.5,
        )
        ax[1].plot(
            [0, nlive],
            [0, 1],
            color="black",
            linewidth=1.25,
            linestyle=":",
            label="cmf",
        )

    ax[0].legend(loc="lower right")
    ax[1].legend(loc="lower right")
    ax[0].set_xlim([0, nlive - 1])
    ax[1].set_xlim([0, nlive - 1])
    ax[0].set_xlabel("Insertion indices")
    ax[1].set_xlabel("Insertion indices")

    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")
        plt.close(fig)
    else:
        return fig


@nessai_style()
def plot_loss(epoch, history, filename=None):
    """
    Plot the loss function per epoch.

    Parameters
    ----------
    epoch : int
        Final training epoch
    history : dict
        Dictionary with keys ``'loss'`` and ``'val_loss'``
    filename : str, optional
        Path for saving the figure. If not specified figure is returned
        instead.
    """
    fig, ax = plt.subplots()
    epochs = np.arange(1, epoch + 1, 1)
    plt.plot(epochs, history["loss"], label="Loss")
    plt.plot(epochs, history["val_loss"], label="Val. loss")
    plt.xlabel("Epochs")
    plt.ylabel("Negative log-likelihood")
    plt.legend()
    plt.tight_layout()
    if any(h < 0 for h in history["loss"]):
        plt.yscale("symlog")
    else:
        plt.yscale("log")

    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")
        plt.close(fig)
    else:
        return fig


@nessai_style()
def plot_trace(
    log_x: np.ndarray,
    nested_samples: np.ndarray,
    parameters: Optional[List[str]] = None,
    live_points: Optional[np.ndarray] = None,
    log_x_live_points: Optional[np.ndarray] = None,
    labels: Optional[List[str]] = None,
    filename: Optional[str] = None,
    **kwargs,
):
    """Produce trace plot for the nested samples.

    By default this includes all parameters in the samples, not just those
    included in the model being sampled.

    Parameters
    ----------
    log_x : array_like
        Array of log prior volumes
    nested_samples : ndarray
        Array of nested samples to plot
    parameters : list, optional
        List of parameters to include the trace plot. If not specified, all of
        the parameters in the nested samples are included.
    live_points : ndarray
        Optional array of live points to include in the plot. See also
        :code:`log_x_live_points`.
    log_x_live_points : ndarray
        Optional array of log-prior volumes for the live points. Required if
        :code:`live_points` is specified.
    labels : list, optional
        List of labels to use instead of the names of parameters
    filename : str, optional
        Filename for saving the plot, if none plot is not saved and figure
        is returned instead.
    kwargs :
        Keyword arguments passed to :code:`matplotlib.pyplot.plot`.
    """
    default_kwargs = dict(
        marker=",",
        linestyle="",
    )

    nested_samples = np.asarray(nested_samples)
    if not nested_samples.dtype.names:
        raise TypeError("Nested samples must be a structured array")

    if parameters is None:
        parameters = nested_samples.dtype.names
    if labels is None:
        labels = parameters

    if not len(labels) == len(parameters):
        raise RuntimeError(
            f"List of labels is the wrong length ({len(labels)}) for the "
            f"parameters: {parameters}."
        )
    if kwargs:
        default_kwargs.update(kwargs)

    figsize = (5, min(config.plotting.max_figsize, 3 * len(labels)))
    fig, axes = plt.subplots(len(labels), 1, figsize=figsize, sharex=True)
    if len(labels) > 1:
        axes = axes.ravel()
    else:
        axes = [axes]

    for i, name in enumerate(parameters):
        axes[i].plot(log_x, nested_samples[name], **default_kwargs)
        if live_points is not None:
            axes[i].plot(
                log_x_live_points, live_points[name], **default_kwargs
            )
        axes[i].set_ylabel(labels[i])

    axes[-1].set_xlabel("log X")
    axes[-1].invert_xaxis()

    if filename is not None:
        try:
            fig.savefig(filename, bbox_inches="tight")
        except ValueError as e:
            logger.warning(f"Could not save trace plot. Error: {e}")
        plt.close(fig)
    else:
        return fig


@nessai_style()
def plot_histogram(samples, label=None, filename=None, **kwargs):
    """Plot a histogram of samples.

    Parameters
    ----------
    samples : array_like
        Samples to plot.
    label : str, optional
        Label to the x axis.
    filename : str, optional
        Filename for saving the plot. If not specified, figure is returned.
    kwargs :
        Keyword arguments passed to `matplotlib.pyplot.hist`.
    """
    default_kwargs = dict(histtype="step", bins=auto_bins(samples))
    default_kwargs.update(kwargs)
    fig = plt.figure()
    plt.hist(samples, **default_kwargs)
    if label is not None:
        plt.xlabel(label)
    if filename is not None:
        fig.savefig(filename, bbox_inches="tight")
        plt.close(fig)
    else:
        return fig


@nessai_style()
def corner_plot(
    array,
    include=None,
    exclude=None,
    labels=None,
    truths=None,
    filename=None,
    **kwargs,
):
    """Produce a corner plot for a structured array.

    Removes any fields with no dynamic range.

    Parameters
    ----------
    array : numpy.ndarray
        Structured array
    include : Optional[list]
        List of parameters to plot.
    exclude : Optional[list]
        List of parameters to exclude.
    labels : Optional[Iterable]
        Labels for each parameter that is to be plotted.
    truths : Optional[Union[Iterable, Dict]]
        Truth values for each parameters, parameters can be skipped by setting
        the value to None.
    filename : Optional[str]
        Filename for saving the plot. If not specified, figure is returned.
    kwargs : Dict[Any]
        Dictionary of keyword arguments passed to :code:`corner.corner`.
    """
    import corner

    default_kwargs = dict(
        bins=32,
        smooth=0.9,
        color=config.plotting.base_colour,
        truth_color=config.plotting.highlight_colour,
        quantiles=[0.16, 0.5, 0.84],
        levels=(1 - np.exp(-0.5), 1 - np.exp(-2), 1 - np.exp(-9 / 2.0)),
        plot_density=True,
        plot_datapoints=True,
        fill_contours=True,
        show_titles=True,
        hist_kwargs=dict(density=True),
    )
    if kwargs:
        default_kwargs.update(kwargs)

    if include and exclude:
        raise ValueError("Cannot specify both `include` and `exclude`")

    if exclude:
        include = [n for n in array.dtype.names if n not in exclude]
    if include:
        array = array[include]
    if labels is None:
        labels = np.asarray(array.dtype.names)
    else:
        labels = np.asarray(labels)

    unstruct_array = live_points_to_array(array)

    has_range = np.array(
        [
            (~np.isnan(v).all()) and (~(np.nanmin(v) == np.nanmax(v)))
            for v in unstruct_array.T
        ],
        dtype=bool,
    )
    if not all(has_range):
        logger.warning(
            "Some parameters have no dynamic range. Removing: "
            f"{[n for n, b in zip(array.dtype.names, has_range) if not b]}"
        )
    unstruct_array = unstruct_array[..., has_range]

    if len(labels) != unstruct_array.shape[-1]:
        labels = labels[has_range]

    if truths is not None:
        if isinstance(truths, dict):
            if include:
                truths = np.array([truths[n] for n in include])
            else:
                truths = np.fromiter(truths.values(), float)
        else:
            truths = np.asarray(truths)
        if len(truths) != unstruct_array.shape[-1]:
            if not all(has_range):
                truths = truths[has_range]
            else:
                raise ValueError(
                    "Length of truths does not match number of "
                    "parameters being plotted"
                )

    fig = corner.corner(
        unstruct_array, truths=truths, labels=labels, **default_kwargs
    )

    if filename is not None:
        try:
            fig.savefig(filename, bbox_inches="tight")
        except ValueError as e:
            logger.warning(f"Could not save corner plot. Error: {e}")
        plt.close(fig)
    else:
        return fig



# -*- coding: utf-8 -*-
"""
Tests for settings utilities.
"""
from unittest.mock import patch

import pytest
from nessai.flowsampler import FlowSampler
from nessai.utils.settings import (
    get_all_kwargs,
    get_run_kwargs_list,
    _get_importance_methods,
    _get_standard_methods,
)


def test_get_standard_methods():
    """Assert a list of methods is returned"""
    kwds, run = _get_standard_methods()
    assert len(kwds) == 4
    assert len(run) == 1


def test_get_importance_methods():
    """Assert a list of methods is returned"""
    kwds, run = _get_importance_methods()
    assert len(kwds) == 3
    assert len(run) == 1


@pytest.mark.parametrize(
    "ins, get_method",
    [(False, "_get_standard_methods"), (True, "_get_importance_methods")],
)
def test_get_all_kwargs(ins, get_method):
    """Assert the correct dictionary is returned.

    Positional arguments should be ignored.
    """

    def func0(a, b, c=2, d=None):
        pass

    def func1(e, f, g=3, h=True):
        pass

    expected = dict(c=2, d=None, g=3, h=True)

    with patch(
        f"nessai.utils.settings.{get_method}",
        return_value=[
            [
                func0,
            ],
            [
                func1,
            ],
        ],
    ):
        out = get_all_kwargs(importance_nested_sampler=ins, split_kwargs=False)

    assert out == expected


@pytest.mark.parametrize(
    "ins, get_method",
    [(False, "_get_standard_methods"), (True, "_get_importance_methods")],
)
def test_get_all_kwargs_split(ins, get_method):
    """Assert the correct dictionary is returned.

    Positional arguments should be ignored.
    """

    def func0(a, b, c=2, d=None):
        pass

    def func1(e, f, g=3, h=True):
        pass

    expected = (dict(c=2, d=None), dict(g=3, h=True))

    with patch(
        f"nessai.utils.settings.{get_method}",
        return_value=[
            [
                func0,
            ],
            [
                func1,
            ],
        ],
    ):
        out = get_all_kwargs(importance_nested_sampler=ins, split_kwargs=True)

    assert out == expected


@pytest.mark.parametrize(
    "ins, run_method",
    [(False, "run_standard_sampler"), (True, "run_importance_nested_sampler")],
)
def test_get_run_kwargs_list(ins, run_method):
    """Assert the correct list is returned"""

    def func(a, b=1, c=None):
        pass

    expected = ["b", "c"]

    with patch.object(FlowSampler, run_method, func):
        out = get_run_kwargs_list(importance_nested_sampler=ins)

    assert out == expected



# -*- coding: utf-8 -*-
"""
Specific reparameterisations for gravitational-wave inference.
"""

import logging

import numpy as np

from ..reparameterisations import (
    default_reparameterisations,
    Reparameterisation,
    RescaleToBounds,
    AnglePair,
    get_reparameterisation,
)

from ..priors import log_uniform_prior

from .utils import get_distance_converter


logger = logging.getLogger(__name__)


def get_gw_reparameterisation(reparameterisation):
    """
    Get a reparameterisation from the default list plus specific GW
    classes.

    Parameters
    ----------
    reparameterisation : str, \
            :obj:`nessai.reparameterisations.Reparameterisation`
        Name of the reparameterisations to return or a class that inherits from
        :obj:`~nessai.reparameterisations.Reparameterisation`

    Returns
    -------
    :obj:`nessai.reparameteristaions.Reparameterisation`
        Reparameterisation class.
    dict
        Keyword arguments for the specific reparameterisation.
    """
    return get_reparameterisation(reparameterisation, defaults=default_gw)


class DistanceReparameterisation(RescaleToBounds):
    """Reparameterisation for distance.

    If the prior is specified and is one of the known priors then a rescaling
    is applied such that the resulting parameter has a uniform prior. If the
    prior is not specified, then the distance is rescaled an inversion is
    allowed on only the upper bound.

    Known priors
    ------------
    * Power-law: requires specifying the power. See converter kwargs.
    * Uniform-comoving-volume: uses a lookup table to convert to co-moving
    distance.

    Parameters
    ----------
    parameters : str
        Name of distance parameter to rescale.
    prior : {'power-law', 'uniform-comoving-volume'}, optional
        Prior used for the distance parameter
    prior_bounds : tuple
        Tuple of lower and upper bounds on the prior
    converter_kwargs : dict, optional
        Keyword arguments parsed to converter object that converts the distance
        to a parameter with a uniform prior.
    allowed_bounds : list, optional
        List of the allowed bounds for inversion
    kwargs :
        Additional kwargs are parsed to the parent class.
    """

    requires_bounded_prior = True

    def __init__(
        self,
        parameters=None,
        allowed_bounds=["upper"],
        allow_both=False,
        converter_kwargs=None,
        prior=None,
        prior_bounds=None,
        **kwargs,
    ):

        if isinstance(parameters, str):
            parameters = [parameters]

        if len(parameters) > 1:
            raise RuntimeError(
                "DistanceReparameterisation only supports one parameter"
            )

        dc_class = get_distance_converter(prior)

        if converter_kwargs is None:
            converter_kwargs = {}
        self.distance_converter = dc_class(
            d_min=prior_bounds[parameters[0]][0],
            d_max=prior_bounds[parameters[0]][1],
            **converter_kwargs,
        )

        pre_rescaling = (
            self.distance_converter.to_uniform_parameter,
            self.distance_converter.from_uniform_parameter,
        )

        super().__init__(
            parameters=parameters,
            prior=prior,
            prior_bounds=prior_bounds,
            pre_rescaling=pre_rescaling,
            **kwargs,
        )

        if self.distance_converter.has_conversion:
            self._prime_prior = log_uniform_prior
            self.has_prime_prior = True
            if not self.distance_converter.has_jacobian:
                logger.debug(
                    "Distance converter does not have Jacobian, "
                    "require prime prior"
                )
                self.requires_prime_prior = True
            self.update_prime_prior_bounds()
        else:
            self.has_prime_prior = False

        self.detect_edges_kwargs["allowed_bounds"] = allowed_bounds
        self.detect_edges_kwargs["allow_both"] = allow_both
        self.detect_edges_kwargs["x_range"] = self.prior_bounds[
            self.parameters[0]
        ]


class DeltaPhaseReparameterisation(Reparameterisation):
    """Reparameterisation that converts phase to delta phase.

    The Jacobian determinant of this transformation is 1.

    Requires "psi" and "theta_jn".

    Parameters
    ----------
    parameters : Union[str, List[str]]
        Name(s) of the parameter(s).
    prior_bounds : Union[list, dict]
        Prior bounds for the parameters
    """

    def __init__(self, parameters=None, prior_bounds=None):
        super().__init__(parameters=parameters, prior_bounds=prior_bounds)
        self.requires = ["psi", "theta_jn"]
        self.prime_parameters = ["delta_phase"]

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space to x'-space.

        Parameters
        ----------
        x : structured array
            Array of inputs
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated

        Returns
        -------
        x, x_prime : structured arrays
            Update version of the x and x_prime arrays
        log_j : array_like
            Updated log Jacobian determinant
        """
        x_prime[self.prime_parameters[0]] = (
            x[self.parameters[0]] + np.sign(np.cos(x["theta_jn"])) * x["psi"]
        )
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated

        Returns
        -------
        x, x_prime : structured arrays
            Update version of the x and x_prime arrays
        log_j : array_like
            Updated log Jacobian determinant
        """
        x[self.parameters[0]] = np.mod(
            x_prime[self.prime_parameters[0]]
            - np.sign(np.cos(x["theta_jn"])) * x["psi"],
            2 * np.pi,
        )
        return x, x_prime, log_j


default_gw = {
    "distance": (
        DistanceReparameterisation,
        {
            "boundary_inversion": True,
            "detect_edges": True,
            "inversion_type": "duplicate",
        },
    ),
    "time": (RescaleToBounds, {"offset": True, "update_bounds": True}),
    "sky-ra-dec": (AnglePair, {"convention": "ra-dec"}),
    "sky-az-zen": (AnglePair, {"convention": "az-zen"}),
    "mass_ratio": (
        RescaleToBounds,
        {
            "detect_edges": True,
            "boundary_inversion": True,
            "inversion_type": "duplicate",
            "update_bounds": True,
        },
    ),
    "mass": (RescaleToBounds, {"update_bounds": True}),
    "delta_phase": (DeltaPhaseReparameterisation, {}),
    "delta-phase": (DeltaPhaseReparameterisation, {}),
}


default_gw.update(default_reparameterisations)



"""Test sampling with the importance nested sampler"""

import os
import pickle

from nessai.flowsampler import FlowSampler
import numpy as np
import pytest


@pytest.mark.slow_integration_test
@pytest.mark.flaky(reruns=3)
@pytest.mark.parametrize("save_log_q", [False, True])
def test_ins_resume(tmp_path, model, flow_config, save_log_q):
    """Assert the INS sampler resumes correctly"""
    output = tmp_path / "test_ins_resume"
    fp = FlowSampler(
        model,
        output=output,
        resume=True,
        nlive=500,
        min_samples=50,
        plot=False,
        flow_config=flow_config,
        importance_nested_sampler=True,
        max_iteration=2,
        save_log_q=save_log_q,
    )
    fp.run()

    original_log_q = fp.ns.log_q.copy()

    assert fp.ns.iteration == 2
    assert os.path.exists(os.path.join(output, "nested_sampler_resume.pkl"))

    fp = FlowSampler(
        model,
        output=output,
        resume=True,
        flow_config=flow_config,
        plot=False,
        importance_nested_sampler=True,
    )
    new_log_q = fp.ns.log_q.copy()

    assert fp.ns.max_iteration == 2
    assert fp.ns.finalised is True
    np.testing.assert_array_almost_equal(new_log_q, original_log_q)


@pytest.mark.slow_integration_test
def test_ins_checkpoint_callback(tmp_path, model, flow_config):
    output = tmp_path / "test_ins_checkpoint_callback"

    filename = os.path.join(output, "test.pkl")
    resume_file = "resume.pkl"

    def checkpoint_callback(state):
        with open(filename, "wb") as f:
            pickle.dump(state, f)

    fs = FlowSampler(
        model,
        output=output,
        resume=True,
        nlive=500,
        min_samples=50,
        plot=False,
        flow_config=flow_config,
        checkpoint_on_iteration=True,
        checkpoint_interval=1,
        importance_nested_sampler=True,
        max_iteration=2,
        resume_file=resume_file,
        checkpoint_callback=checkpoint_callback,
    )
    fs.run()
    assert fs.ns.iteration == 2
    assert os.path.exists(filename)
    assert not os.path.exists(os.path.join(output, resume_file))

    del fs

    with open(filename, "rb") as f:
        resume_data = pickle.load(f)

    resume_data.test_variable = "abc"

    fs = FlowSampler(
        model,
        output=output,
        resume=True,
        nlive=500,
        min_samples=50,
        plot=False,
        flow_config=flow_config,
        checkpoint_on_iteration=True,
        checkpoint_interval=1,
        importance_nested_sampler=True,
        max_iteration=2,
        checkpoint_callback=checkpoint_callback,
        resume_data=resume_data,
        resume_file=resume_file,
    )
    assert fs.ns.iteration == 2
    assert fs.ns.finalised is True
    assert fs.ns.test_variable == "abc"



"""General configuration for the test suite"""

import sys

from numpy.random import seed
import numpy as np
import pytest
from scipy.stats import norm
import time
import torch
import multiprocessing

from nessai.livepoint import (
    add_extra_parameters_to_live_points,
    reset_extra_live_points_parameters,
)
from nessai.model import Model
from nessai.utils.testing import IntegrationTestModel


seed(170817)
torch.manual_seed(170817)

_requires_dependency_cache = dict()


@pytest.fixture()
def model():
    class TestModel(Model):
        def __init__(self):
            self.bounds = {"x": [-5, 5], "y": [-5, 5]}
            self.names = ["x", "y"]

        def log_prior(self, x):
            log_p = np.log(self.in_bounds(x), dtype="float")
            for n in self.names:
                log_p -= self.bounds[n][1] - self.bounds[n][0]
            return log_p

        def log_likelihood(self, x):
            log_l = 0
            for pn in self.names:
                log_l += norm.logpdf(x[pn])
            return log_l

        def to_unit_hypercube(self, x):
            x_out = x.copy()
            for n in self.names:
                x_out[n] = (x[n] - self.bounds[n][0]) / np.ptp(self.bounds[n])
            return x_out

        def from_unit_hypercube(self, x):
            x_out = x.copy()
            for n in self.names:
                x_out[n] = np.ptp(self.bounds[n]) * x[n] + self.bounds[n][0]
            return x_out

    return TestModel()


@pytest.fixture()
def integration_model():
    return IntegrationTestModel()


@pytest.fixture()
def flow_config():
    d = dict(
        max_epochs=5,
        model_config=dict(
            n_blocks=2,
            n_neurons=2,
            n_layers=1,
            device_tag="cpu",
            kwargs=dict(batch_norm_between_layers=False),
        ),
    )
    return d


@pytest.fixture()
def wait():
    """Sleep for 0.01s for timing tests.

    Time resolution on Windows is less than Linux and macOS
    """

    def func(*args, **kwargs):
        time.sleep(0.01)

    return func


@pytest.fixture(params=["fork", "forkserver", "spawn"])
def mp_context(request):
    """Multiprocessing context to test"""
    if sys.platform == "win32" and request.param != "spawn":
        pytest.skip("Windows only supports the 'spawn' start method")
    return multiprocessing.get_context(request.param)


@pytest.fixture()
def ins_parameters():
    """Add (and remove) the standard INS parameters for the tests."""
    # Before every test
    add_extra_parameters_to_live_points(["logQ", "logW"])
    yield
    reset_extra_live_points_parameters()


def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line(
        "markers", "cuda: mark test to indicate it requires CUDA"
    )
    config.addinivalue_line(
        "markers",
        "requires(package): mark test to only run if the package can be "
        "imported",
    )
    config.addinivalue_line(
        "markers",
        "skip_on_windows: mark test to indicated it should be skipped on "
        "Windows",
    )
    config.addinivalue_line(
        "markers",
        "bilby_compatibility: mark test as a bilby compatibility test, these "
        "tests will be skipped unless the command line option is specified.",
    )


def pytest_addoption(parser):
    parser.addoption(
        "--bilby-compatibility",
        action="store_true",
        default=False,
        help="Run bilby compatibility tests",
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--bilby-compatibility"):
        return
    skip_bilby_compat = pytest.mark.skip(
        reason="Need --bilby-compatibility to run"
    )
    for item in items:
        if "bilby_compatibility" in item.keywords:
            item.add_marker(skip_bilby_compat)


def pytest_runtest_setup(item):
    for mark in item.iter_markers(name="cuda"):
        if not torch.cuda.is_available():
            pytest.skip("Test requires CUDA")
    for mark in item.iter_markers(name="requires"):
        name = mark.args[0]
        if name in _requires_dependency_cache:
            skip_it = _requires_dependency_cache[name]
        else:
            try:
                __import__(name)
                skip_it = False
            except ImportError:
                skip_it = True

            _requires_dependency_cache[name] = skip_it

        reason = "Missing dependency: {}".format(name)
        if skip_it:
            pytest.skip(reason)
    for mark in item.iter_markers(name="skip_on_windows"):
        if sys.platform == "win32":
            pytest.skip("Test does not run on Windows")



"""Tests related to sampling from ImportanceFlowProposal"""

from unittest.mock import MagicMock, create_autospec

from nessai.flowmodel.importance import ImportanceFlowModel
from nessai.livepoint import numpy_array_to_live_points, live_points_to_array
from nessai.proposal.importance import ImportanceFlowProposal as IFP
import numpy as np
from scipy.special import logsumexp
import pytest


@pytest.mark.parametrize("n", [1, 10])
@pytest.mark.usefixtures("ins_parameters")
def test_draw_from_prior(ifp, x, n):
    """Test drawing from the prior"""
    n_proposals = 4
    x_prime = np.random.randn(n, 2)
    log_j = np.random.rand(n)
    log_Q = np.random.randn(n)
    log_q = np.random.randn(n_proposals, n)
    ifp.model.sample_unit_hypercube = MagicMock(return_value=x)
    ifp.rescale = MagicMock(return_value=(x_prime, log_j))
    ifp.compute_log_Q = MagicMock(return_value=(log_Q, log_q))

    x_out, log_q_out = IFP.draw_from_prior(ifp, n)
    assert log_q_out.shape == (n_proposals, n)

    assert x_out is x


@pytest.mark.parametrize("n", [1, 3, 10, 100])
@pytest.mark.parametrize("test_counts", [False, True])
def test_draw_from_flows(ifp, model, n, test_counts):

    n_flows = 5
    weights = np.random.rand(n_flows + 1)
    if test_counts:
        counts = np.random.multinomial(n, weights / weights.sum())
        print(counts)
    else:
        counts = None
    ifp.n_proposals = n_flows + 1

    def sample_ith(it, N):
        return np.random.rand(N, model.dims)

    def to_prime(x):
        return x.copy(), np.zeros(x.size)

    def inverse_rescale(x):
        x = numpy_array_to_live_points(x, model.names)
        return x, np.zeros(x.size)

    def rescale(x):
        x = live_points_to_array(x, model.names)
        return x, np.zeros(x.shape[0])

    def log_prob_all(x):
        return np.log(np.random.rand(len(x), n_flows))

    ifp.model = model
    ifp.flow = create_autospec(ImportanceFlowModel)
    ifp.flow.sample_ith = MagicMock(side_effect=sample_ith)
    ifp.flow.log_prob_all = MagicMock(side_effect=log_prob_all)
    ifp.to_prime = MagicMock(side_effect=to_prime)
    ifp.inverse_rescale = MagicMock(side_effect=inverse_rescale)
    ifp.rescale = MagicMock(side_effect=rescale)

    x, log_q, actual_counts = IFP.draw_from_flows(
        ifp, n, weights=weights, counts=counts
    )
    assert len(x) == n
    assert log_q.shape == (n, n_flows + 1)
    assert all(log_q[:, 0] == 0)

    assert np.isfinite(x["logP"]).all()
    assert np.isnan(x["logL"]).all()

    assert np.sum(actual_counts) == n
    if test_counts:
        np.testing.assert_array_equal(actual_counts, counts)


@pytest.mark.usefixtures("ins_parameters")
def test_draw(ifp, model):

    n_proposals = 5
    n_draw = 100
    ifp.n_proposals = n_proposals
    ifp.level_count = n_proposals - 1
    ifp.model = model
    ifp.dtype = model.new_point().dtype
    ifp._weights = {-1: 0.2, 0: 0.2, 2: 0.2, 3: 0.4, 4: np.nan}
    ifp.weights_array = np.fromiter(ifp._weights.values(), float)

    def inverse_rescale(x):
        x = numpy_array_to_live_points(x, model.names)
        return model.from_unit_hypercube(x), np.zeros(x.size)

    def rescale(x):
        x = model.to_unit_hypercube(x)
        x = live_points_to_array(x, model.names)
        return x, np.zeros(x.shape[0])

    def sample_ith(i, N):
        assert i == (n_proposals - 1)
        return np.random.rand(N, model.dims)

    def compute_log_Q(x_prime, log_j=None, n=None):
        log_q = (
            np.log(np.random.rand(len(x_prime), n_proposals)) + log_j[:, None]
        )
        log_Q = logsumexp(log_q, b=ifp.weights_array, axis=1)
        return log_Q, log_q

    ifp.rescale = rescale
    ifp.inverse_rescale = inverse_rescale
    ifp.compute_log_Q = compute_log_Q
    ifp.flow = create_autospec(ImportanceFlowModel)
    ifp.flow.sample_ith = sample_ith

    samples_out, log_q_out = IFP.draw(ifp, n_draw)

    assert len(samples_out) == n_draw
    assert log_q_out.shape == (n_draw, n_proposals)
